from enum import Enum


class Order(Enum):
    ASCENDING = "ascending"
    DESCENDING = "descending"


class Sort():
    """
    **Specification of the sort order**
    """

    def __init__(self, key: str, order: Order) -> None:
        self._sort = {
            "key": key,
            "order": order.value
        }



from __future__ import annotations

from aperturedb.Entities import Entities


class Blobs(Entities):
    """
    **The object mapper representation of blobs in ApertureDB.**

    This class is a layer on top of the native query.
    It facilitates interactions with blobs in the database in the pythonic way.
    """
    db_object = "_Blob"



import logging
import random

from aperturedb.Connector import Connector
from aperturedb.ParallelQuery import ParallelQuery
from aperturedb.Subscriptable import Subscriptable

logger = logging.getLogger(__name__)

# Tests for parallel which don't involve data.


class GeneratorWithErrors(Subscriptable):
    def __init__(self, commands_per_query=1, elements=100, error_pct=.5) -> None:
        super().__init__()
        self.commands_per_query = commands_per_query
        self.elements = elements
        self.error_pct = error_pct

    def __len__(self):
        return self.elements

    def getitem(self, subscript):
        query = []
        blobs = []
        for i in range(self.commands_per_query):
            if random.randint(0, 100) <= (self.error_pct * 100):
                query.append({
                    "BadCommand": {
                    }
                })
            else:
                query.append({
                    "FindEntity": {
                        "results": {
                            "count": True
                        }
                    }
                })

        return query, blobs


class TestParallel():
    """
    These check operation of ParallelQuery
    """

    def test_someBadQueries(self, db: Connector):
        """
        Verifies that it handles some queries returning errors
        """
        try:
            elements = 100
            generator = GeneratorWithErrors(elements=elements)
            querier = ParallelQuery(db, dry_run=False)
            querier.query(generator, batchsize=2,
                          numthreads=8,
                          stats=True)
            logger.info(querier.get_succeeded_commands())
            assert querier.get_succeeded_commands() < elements
        except Exception as e:
            print(e)
            print("Failed to renew Session")
            assert False

    def test_allBadQueries(self, db: Connector):
        """
        Verifies that it handles all queries returning errors
        """
        try:
            elements = 100
            generator = GeneratorWithErrors(elements=elements, error_pct=1)
            querier = ParallelQuery(db, dry_run=False)
            querier.query(generator, batchsize=2,
                          numthreads=8,
                          stats=True)
            logger.info(querier.get_succeeded_commands())
            assert querier.get_succeeded_commands() == 0
        except Exception as e:
            print(e)
            print("Failed to renew Session")
            assert False



from aperturedb.DescriptorDataCSV import DescriptorDataCSV
from aperturedb.ParallelLoader import ParallelLoader
from io import BytesIO, TextIOWrapper
import sys

# stats had some issues with displaying of computed data.
# These tests reproduce the issue.
# Run it with: (to avoid pytest's output)
# pytest  test_Stats.py -s --no-summary


class TestStats():
    def ingest_with_capture(self, data, db):
        loader = ParallelLoader(db)
        # setup the environment
        old_stdout = sys.stdout
        sys.stdout = TextIOWrapper(BytesIO(), sys.stdout.encoding)

        loader.ingest(data, batchsize=99, numthreads=31, stats=True)
        sys.stdout.seek(0)      # jump to the start
        out = sys.stdout.read()  # read output

        sys.stdout = old_stdout
        return out

    def validate_stats(self, out, assertions):
        for line in out.splitlines():
            if ":" in line:
                stats = line.split(":")
                if len(stats) == 2:
                    first, second = line.split(":")
                    print(first, second)
                    if first in assertions:
                        assert assertions[first.strip()](second.strip()) == True, \
                            f"Assertion failed for '{first}' with value {second}"

    def test_stats_all_errors_non_equal_last_batch(self, db, utils):
        utils.remove_all_objects()
        # Try to ingest descriptors, with no descriptor set, so all queries fail
        data = DescriptorDataCSV(
            "./input/setA.adb.csv", blobs_relative_to_csv=True)
        out = self.ingest_with_capture(data, db)
        assertions = {
            "Total inserted elements": lambda x: float(x) == 0,
            "Overall insertion throughput (element/s)": lambda x: x == "NaN",
        }
        self.validate_stats(out, assertions)



"""
Miscellaneous utility functions for ApertureDB.
"""
import logging
import json
import os
import importlib
import sys
from typing import List

from graphviz import Source, Digraph

from aperturedb.Connector import Connector
from aperturedb.ConnectorRest import ConnectorRest
from aperturedb import ProgressBar
from aperturedb.ParallelQuery import execute_batch
from aperturedb.Configuration import Configuration
from aperturedb.cli.configure import ls
from aperturedb.Query import QueryBuilder

logger = logging.getLogger(__name__)

DESCRIPTOR_CLASS = "_Descriptor"
DESCRIPTOR_CONNECTION_CLASS = "_DescriptorSetToDescriptor"

DEFAULT_METADATA_BATCH_SIZE = 100_000


def import_module_by_path(filepath):
    """
    This function imports a module given a path to a python file.
    """
    module_name = os.path.basename(filepath)[:-3]
    spec = importlib.util.spec_from_file_location(module_name, filepath)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def __create_connector(configuration: Configuration):
    if configuration.use_rest:
        connector = ConnectorRest(
            host=configuration.host,
            port=configuration.port,
            user=configuration.username,
            password=configuration.password,
            config=configuration)
    else:
        connector = Connector(
            host=configuration.host,
            port=configuration.port,
            user=configuration.username,
            password=configuration.password,
            config=configuration)
    logger.info(f"Connected Using: {configuration}")
    return connector


def create_connector():
    """
    **Create a connector to the database.**

    This function reads the active configuration.
    See :ref:`adb config <adb_config>`_ command-line tool for more information.

    Args:
        None

    Returns:
        Connector: The connector to the database.
    """
    all_configs = ls(log_to_console=False)

    ac = all_configs["active"]
    config = None
    if ac is not None:
        # check if the active config is in the global or local
        # Local should be the final choice
        if "global" in all_configs and ac in all_configs["global"]:
            config = all_configs["global"][ac]
        if "local" in all_configs and ac in all_configs["local"]:
            config = all_configs["local"][ac]
    assert config is not None, "No active configuration found."

    env_config = os.environ.get("APERTUREDB_CONFIG")
    if env_config is not None:
        # TODO test me.
        config = all_configs["global"][env_config] if env_config in all_configs["global"] else all_configs["local"][env_config]
        return __create_connector(config)
    # Then check if the local config has active
    else:
        return __create_connector(config)


class Utils(object):
    """
    **A bunch of helper methods to get information from aperturedb.**

    Args:
        object (Connector): The underlying Connector.
    """

    def __init__(self, connector: Connector, verbose=False):
        self.connector = connector.create_new_connection()
        self.verbose = verbose

    def __repr__(self):
        return self.status()

    def print(self, str):
        if self.verbose:
            print(str)

    def execute(self, query, blobs=[], success_statuses=[0]):
        """
        Execute a query.

        Args:
            query (list): The query to execute.
            blobs (list, optional): The blobs to send with the query.
            success_statuses (list, optional): The list of success statuses.

        Returns:
            result: The result of the query.
            blobs: The blobs returned by the query.
        """
        try:
            rc, r, b = execute_batch(
                query, blobs, self.connector, success_statuses=success_statuses)
        except BaseException as e:
            logger.error(self.connector.get_last_response_str())
            raise e

        if rc != 0:
            raise Exception(f"query failed with result {rc}")

        return r, b

    def status(self):
        """
        Executes a `GetStatus` query.
        See :ref:`GetStatus <GetStatus>`_ in the ApertureDB documentation for more information.
        """

        q = [{"GetStatus": {}}]

        self.execute(q)

        return self.connector.get_last_response_str()

    def print_schema(self, refresh=False):

        if refresh:
            logger.warning("get_schema: refresh no longer needed.")
            logger.warning("get_schema: refresh will be deprecated.")
            logger.warning("Please remove 'refresh' parameter.")

        self.get_schema()
        self.connector.print_last_response()

    def get_schema(self, refresh=False):
        """
        Get the schema of the database.
        See :ref:`GetSchema <GetSchema>`_ in the ApertureDB documentation for more information.
        """

        if refresh:
            logger.warning("get_schema: refresh no longer needed.")
            logger.warning("get_schema: refresh will be deprecated.")
            logger.warning("Please remove 'refresh' parameter.")

        query = [{
            "GetSchema": {
            }
        }]

        res, _ = self.connector.query(query)

        schema = {}

        try:
            schema = res[0]["GetSchema"]
        except BaseException as e:
            logger.error(self.connector.get_last_response_str())
            raise e

        return schema

    def visualize_schema(self, filename: str = None, format: str = "png") -> Source:
        """
        Visualize the schema of the database.  Optionally save the visualization to a file in the specified format.

        The returned object can be rendered to a file as follows:
        ```python
        s = utils.visualize_schema()
        s.render("schema", format="png")
        ```

        It can also be displayed inline in a Jupyter notebook:
        ```python
        from IPython.display import display
        s = utils.visualize_schema()
        display(s)
        ```

        Relies on graphviz to be installed.

        Args:
            filename (str, optional): The filename to save the visualization to. Default is None.
            format (str, optional): The format to save the visualization to. Default is "png".

        Returns:
            source: The visualization of the schema.
        """
        r = self.get_schema()

        dot = Digraph(comment='ApertureDB Schema Diagram', node_attr={
                      'shape': 'none', 'fontcolor': '#E2E0F1'}, graph_attr={'rankdir': 'LR'}, edge_attr={'color': '#3A3B9C'})

        # Add entities as nodes and connections as edges
        entities = r['entities']['classes']
        connections = r['connections']['classes']

        for entity, data in entities.items():
            matched = data["matched"]
            # dictionary from name to (matched, indexed, type)
            properties = data["properties"]
            table = f'''<
            <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">
            <TR><TD BGCOLOR="#2A2E78" COLSPAN="3"><B>{entity}</B> ({matched:,})</TD></TR>
            '''
            for prop, (matched, indexed, typ) in properties.items():
                table += f'<TR><TD BGCOLOR="#337EC0"><B>{prop.strip()}</B></TD> <TD BGCOLOR="#337EC0">{matched:,}</TD> <TD BGCOLOR="#337EC0">{"Indexed" if indexed else "Unindexed"}, {typ}</TD></TR>'
            for connection, data in connections.items():
                if data['src'] == entity:
                    matched = data["matched"]
                    # dictionary from name to (matched, indexed, type)
                    properties = data["properties"]
                    table += f'<TR><TD BGCOLOR="#5956F1" COLSPAN="3" PORT="{connection}"><B>{connection}</B> ({matched:,})</TD></TR>'
                    if properties:
                        for prop, (matched, indexed, typ) in properties.items():
                            table += f'<TR><TD BGCOLOR="#33E1FF"><B>{prop.strip()}</B></TD> <TD BGCOLOR="#33E1FF">{matched:,}</TD> <TD BGCOLOR="#33E1FF">{"Indexed" if indexed else "Unindexed"}, {typ}</TD></TR>'
            table += '</TABLE>>'
            dot.node(entity, label=table)

        for connection, data in connections.items():
            dot.edge(f'{data["src"]}:{connection}',
                     f'{data["dst"]}:{connection}')

        # Render the diagram inline
        s = Source(dot.source, filename="schema_diagram.gv", format="png")

        if filename is not None:
            s.render(filename, format=format)

        return s

    def _object_summary(self, name, object):

        total_elements = object["matched"]

        print(f"{name.ljust(20)}")
        if "src" in object:
            print(f"  {object['src']} ====> {object['dst']}")

        print(f"  Total elements: {total_elements}")

        p = object["properties"]

        if p is None:
            return total_elements

        max = 0
        for k in p:
            max = len(k) if max < len(k) else max
        max += 1

        for k in p:
            i = w = " "
            i = "I" if p[k][1] else i

            # Warning if there are some properties not present in all elements
            w = "!" if p[k][0] != total_elements else w
            # Warning if there is a property with "id" in the name not indexes
            w = "!" if "id" in k and not p[k][1] else w
            print(f"{i} {w} {p[k][2].ljust(8)} |"
                  f" {k.ljust(max)} | {str(p[k][0]).rjust(9)} "
                  f"({int(p[k][0]/total_elements*100.0)}%)")

        return total_elements

    def summary(self):
        """
        Print a summary of the database.

        This is essentially a call to :ref:`GetSchema <GetSchema>`_, with the results formatted in a more human-readable way.
        """
        r = self.get_schema()
        s = json.loads(self.status())[0]["GetStatus"]
        version = s["version"]
        status = s["status"]
        info = s["info"]

        if r["entities"] == None:
            total_entities = 0
            entities_classes = []
        else:
            total_entities = r["entities"]["returned"]
            entities_classes = [c for c in r["entities"]["classes"]]

        if r["connections"] == None:
            total_connections = 0
            connections_classes = []
        else:
            total_connections = r["connections"]["returned"]
            connections_classes = [c for c in r["connections"]["classes"]]

        print(f"================== Summary ==================")
        print(f"Database: {self.connector.host}")
        print(f"Version: {version}")
        print(f"Status:  {status}")
        print(f"Info:    {info}")
        print(f"------------------ Entities -----------------")
        print(f"Total entities types:    {total_entities}")
        total_nodes = 0
        for c in entities_classes:
            total_nodes += self._object_summary(c, r["entities"]["classes"][c])

        print(f"---------------- Connections ----------------")
        print(f"Total connections types: {total_connections}")
        total_edges = 0
        for c in connections_classes:
            total_edges += self._object_summary(c,
                                                r["connections"]["classes"][c])

        print(f"------------------ Totals -------------------")
        print(f"Total nodes: {total_nodes}")
        print(f"Total edges: {total_edges}")
        print(f"=============================================")

    def _create_index(self, index_type, class_name, property_key):

        q = [{
            "CreateIndex": {
                "index_type":    index_type,
                "class":         class_name,
                "property_key":  property_key,
            }
        }]

        try:
            self.execute(q, success_statuses=[0, 2])
        except:
            return False

        return True

    def _remove_index(self, index_type, class_name, property_key):

        q = [{
            "RemoveIndex": {
                "index_type":   index_type,
                "class":        class_name,
                "property_key": property_key,
            }
        }]

        try:
            self.execute(q)
        except:
            return False

        return True

    def create_entity_index(self, class_name, property_key, property_type=None):
        """
        Create an index for an entity class.
        See :ref:`CreateIndex <CreateIndex>`_ in the ApertureDB documentation for more information.
        """

        if property_type is not None:
            logger.warning(f"create_entity_index ignores 'property_type'")
            logger.warning(f"'property_type' will be deprecated in the future")
            logger.warning(f"Please remove 'property_type' parameter")

        return self._create_index("entity", class_name, property_key)

    def create_connection_index(self, class_name, property_key, property_type=None):
        """
        Create an index for a connection class.
        See :ref:`CreateIndex <CreateIndex>`_ in the ApertureDB documentation for more information.
        """

        if property_type is not None:
            logger.warning(f"create_connection_index ignores 'property_type'")
            logger.warning(f"'property_type' will be deprecated in the future")
            logger.warning(f"Please remove 'property_type' parameter")

        return self._create_index("connection", class_name, property_key)

    def remove_entity_index(self, class_name, property_key):
        """
        Remove an index for an entity class.
        See :ref:`RemoveIndex <RemoveIndex>`_ in the ApertureDB documentation for more information.
        """

        return self._remove_index("entity", class_name, property_key)

    def remove_connection_index(self, class_name, property_key):
        """
        Remove an index for a connection class.
        See :ref:`RemoveIndex <RemoveIndex>`_ in the ApertureDB documentation for more information.
        """
        return self._remove_index("connection", class_name, property_key)

    def count_images(self, constraints={}) -> int:
        """
        Count the number of images in the database.

        Args:
            constraints (dict, optional): The constraints to apply to the query.
                See the [`Constraints`](../parameter_wrappers/Constraints) wrapper class for more information.

        Returns:
            count: The number of images in the database.
        """
        q = [{
            "FindImage": {
                "blobs": False,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindImage"]["constraints"] = constraints

        res, _ = self.execute(q)
        total_images = res[0]["FindImage"]["count"]

        return total_images

    def count_bboxes(self, constraints=None) -> int:
        """
        Count the number of bounding boxes in the database.

        Args:
            constraints (dict, optional): The constraints to apply to the query.
                See the [`Constraints`](../parameter_wrappers/Constraints) wrapper class for more information.

        Returns:
            count: The number of bounding boxes in the database.
        """
        # The default params in python functions should not be
        # mutable objects.
        # It can lead to https://docs.python-guide.org/writing/gotchas/#mutable-default-arguments
        q = [{
            "FindBoundingBox": {
                "blobs": False,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindBoundingBox"]["constraints"] = constraints

        res, _ = self.execute(q)
        total_connections = res[0]["FindBoundingBox"]["count"]

        return total_connections

    def count_entities(self, entity_class, constraints=None) -> int:
        """
        Count the number of entities in the database.

        Args:
            constraints (dict, optional): The constraints to apply to the query.
                See the [`Constraints`](../parameter_wrappers/Constraints) wrapper class for more information.

        Returns:
            count: The number of entities in the database.
        """
        params = {"results": {"count": True}}
        if constraints:
            params["constraints"] = constraints
        q = QueryBuilder.find_command(entity_class, params=params)
        res, _ = self.execute(query=[q])
        total_entities = res[0][list(q.keys())[0]]["count"]

        return total_entities

    def count_connections(self, connections_class, constraints=None) -> int:
        """
        Count the number of connections in the database.

        Args:
            constraints (dict, optional): The constraints to apply to the query.
                See the [`Constraints`](../parameter_wrappers/Constraints) wrapper class for more information.

        Returns:
            count: The number of connections in the database.
        """

        q = [{
            "FindConnection": {
                "with_class": connections_class,
                "results": {
                    "count": True,
                }
            }
        }]

        if constraints:
            q[0]["FindConnection"]["constraints"] = constraints

        res, _ = self.execute(q)
        total_connections = res[0]["FindConnection"]["count"]

        return total_connections

    def add_descriptorset(self, name: str, dim: int, metric="L2", engine="FaissFlat") -> bool:
        """
        Add a descriptor set to the database.

        Args:
            name (str): The name of the descriptor set.
            dim (int): The dimension of the descriptors.
            metric (str, optional): The metric to use for the descriptors.
            engine (str, optional): The engine to use for the descriptors.

        Returns:
            success (bool): True if the operation was successful, False otherwise.
        """
        query = [{
            "AddDescriptorSet": {
                "name":       name,
                "dimensions": dim,
                "metric":     metric,
                "engine":     engine
            }
        }]

        try:
            self.execute(query)
        except:
            return False

        return True

    def count_descriptorsets(self) -> int:
        """
        Count the number of descriptor sets in the database.

        Returns:
            count: The number of descriptor sets in the database.
        """
        q = [{
            "FindDescriptorSet": {
                "results": {
                    "count": True,
                }
            }
        }]

        res, _ = self.execute(q)
        total_descriptor_sets = res[0]["FindDescriptorSet"]["count"]

        return total_descriptor_sets

    def get_descriptorset_list(self) -> List[str]:
        """
        Get the list of descriptor sets in the database.

        Returns:
            sets (list of str): The list of descriptor sets in the database.
        """
        q = [{
            "FindDescriptorSet": {
                "results": {
                    "list": ["_name"],
                }

            }
        }]

        sets = []
        res, _ = self.execute(q)
        if res[0]["FindDescriptorSet"]["returned"] > 0:
            sets = [ent["_name"]
                    for ent in res[0]["FindDescriptorSet"]["entities"]]

        return sets

    def remove_descriptorset(self, set_name: str) -> bool:
        """
        Remove a descriptor set from the database.

        See :ref:`DeleteDescriptorSet <DeleteDescriptorSet>`_ in the ApertureDB documentation for more information.

        Args:
            set_name (str): The name of the descriptor set.

        Returns:
            success (bool): True if the operation was successful, False otherwise.
        """
        q = [{
            "FindDescriptorSet": {
                "_ref": 1,
                "with_name": set_name,

            }
        }, {
            "DeleteDescriptorSet": {
                "ref": 1
            }
        }]

        try:
            self.execute(q)
        except:
            return False

        return True

    def _remove_objects(self, type, class_name, batch_size):

        if type == "entities":
            cmd = "Entity"
            count = self.count_entities(class_name)
        elif type == "connections":
            cmd = "Connection"
            count = self.count_connections(class_name)
        else:
            raise ValueError("Type must be either 'entities' or 'connections'")

        total = count

        pb = ProgressBar.ProgressBar()

        find = "Find" + cmd
        dele = "Delete" + cmd
        while count > 0:

            q = [{
                find: {
                    "_ref": 1,
                    "with_class": class_name,
                    "results": {
                        "limit": batch_size
                    }
                }
            }, {
                dele: {
                    "ref": 1
                }
            }]

            try:
                self.execute(q)
            except:
                return False

            self.print(
                f"Delete transaction of {batch_size} elements took: {self.connector.get_last_query_time()}")

            count -= batch_size

            if self.verbose:
                pb.update(abs((count - total) / total))

        return True

    def remove_entities(self, class_name: str, batched: bool = False, batch_size: int = 10000) -> bool:
        """
        Remove all entities of a given class from the database.

        See :ref:`DeleteEntity <DeleteEntity>`_ in the ApertureDB documentation for more information.

        Args:
            class_name (str): The class of the entities to remove.
            batched (bool, optional): Whether to batch the operation. Default is False.
            batch_size (int, optional): The batch size to use. Default is 10000.

        Returns:
            success (bool): True if the operation was successful, False otherwise.
        """
        # We used to batch because removing a large number of object did not work ok.
        # Now it seems to work ok, so we don't batch anymore.
        # We keep the option in case we need to batch, but not default.
        if batched:
            return self._remove_objects("entities", class_name, batch_size)

        query = [{
            "DeleteEntity": {
                "with_class": class_name
            }
        }]

        try:
            r, _ = self.execute(query)
            count = r[0]["DeleteEntity"]["count"]
            self.print(f"Deleted {count} entities")
            self.print(
                f"Delete transaction took: {self.connector.get_last_query_time()}")
        except:
            logger.error("Could not get count of deleted entities")
            return False

        return True

    def remove_connections(self, class_name, batched=False, batch_size=10000):
        """
        Remove all connections of a given class from the database.

        See :ref:`DeleteConnection <DeleteConnection>`_ in the ApertureDB documentation for more information.

        Args:
            class_name (str): The class of the connections to remove.
            batched (bool, optional): Whether to batch the operation. Default is False.
            batch_size (int, optional): The batch size to use. Default is 10000.

        Returns:
            success (bool): True if the operation was successful, False otherwise.
        """
        # We used to batch because removing a large number of object did not work ok.
        # Now it seems to work ok, so we don't batch anymore.
        # We keep the option in case we need to batch, but not default.
        if batched:
            return self._remove_objects("connections", class_name, batch_size)

        query = [{
            "DeleteConnection": {
                "with_class": class_name
            }
        }]

        try:
            r, _ = self.execute(query)
            count = r[0]["DeleteConnection"]["count"]
            self.print(f"Deleted {count} connections")
            self.print(
                f"Delete transaction took: {self.connector.get_last_query_time()}")
        except:
            logger.error("Could not get count of deleted connections")
            return False

        return True

    def remove_all_descriptorsets(self) -> bool:
        """
        Remove all descriptor sets from the database, together with descriptors, indexes and connections.

        Returns:
            success (bool): True if the operation was successful, False otherwise.
        """
        self.print("Removing indexes...")

        idx_props = self.get_indexed_props(DESCRIPTOR_CLASS)

        for idx in idx_props:
            self.remove_entity_index(DESCRIPTOR_CLASS, idx)

        self.print("Done removing indexes.")

        self.print("Removing connections...")
        if not self.remove_connections(DESCRIPTOR_CONNECTION_CLASS):
            print("Error removing connections.")
            return False
        self.print("Done removing connections.")

        self.print("Removing descriptors...")
        if not self.remove_entities(DESCRIPTOR_CLASS):
            print("Error removing descriptors.")
            return False
        self.print("Done removing descriptors.")

        sets = self.get_descriptorset_list()

        self.print("Removing sets...")
        for s in sets:
            self.print("Removing {}...".format(s))
            self.remove_descriptorset(s)

        self.print("Done removing sets.")

    def get_indexed_props(self, class_name: str, type="entities") -> List[str]:
        """
        Returns all the indexed properties for a given class.

        Args:
            class_name (str): The class name.
            type (str, optional): The type of the class. Default is "entities".

        Returns:
            indexed_props (list of str): The list of indexed properties.
        """

        if type not in ["entities", "connections"]:
            raise ValueError("Type must be either 'entities' or 'connections'")

        schema = self.get_schema()

        try:
            indexed_props = schema[type]["classes"][class_name]["properties"]
            indexed_props = [
                k for k in indexed_props.keys() if indexed_props[k][1]]
        except:
            indexed_props = []

        return indexed_props

    def count_descriptors_in_set(self, set_name: str) -> int:
        """
        Count the number of descriptors in a descriptor set.

        Args:
            set_name (str): The name of the descriptor set.

        Returns:
            total (int): The number of descriptors in the set.
        """
        total = -1

        q = [{
            "FindDescriptorSet": {
                "_ref": 1,
                "with_name": set_name,
            }
        }, {
            "FindDescriptor": {
                "set": set_name,
                "is_connected_to": {
                    "ref": 1,
                    "connection_class": "_DescriptorSetToDescriptor"
                },
                "results": {
                    "count": True
                }
            }
        }]

        res, _ = self.execute(q)
        total = res[1]["FindDescriptor"]["count"]

        return total

    def remove_all_indexes(self) -> bool:
        """Remove all indexes from the database.

        This may improve the performance of remove_all_objects.
        It may improve or degrade the performance of other operations.

        Note that this only removes populated indexes.

        Returns:
            success (bool): True if the operation was successful, False otherwise.
        """
        def find_indexes(schema):
            typemap = dict(entities="entity", connections="connection")
            for typ, data in schema['GetSchema'].items():
                if typ in typemap and type(data) == dict and 'classes' in data:
                    for clas, classdata in data['classes'].items():
                        if 'properties' in classdata and classdata['properties']:
                            for property_key, (_, indexed, _) in classdata['properties'].items():
                                if indexed:
                                    yield {"index_type": typemap[typ], "class": clas, "property_key": property_key}

        try:
            r, _ = self.execute([{"GetSchema": {}}])
            schema = r[0]
            indexes = list(find_indexes(schema))
            query = [{"RemoveIndex": index} for index in indexes]
            query.append({"GetSchema": {}})
            r2, _ = self.execute(query)
            schema2 = r2[-1]
            indexes2 = list(find_indexes(schema2))
            if indexes2:
                logger.error(f"Failed to remove all indexes: {indexes2}")
                return False
        except BaseException as e:
            logger.exception(e)
            return False

        return True

    def remove_all_objects(self) -> bool:
        """
        Remove all objects from the database.

        This includes images, videos, blobs, clips, descriptor sets, descriptors, bounding boxes, polygons, frames, entities, connections, indexes and connections.

        Returns:
            success (bool): True if the operation was successful, False otherwise.
        """
        cmd = {"constraints": {"_uniqueid": ["!=", "0.0.0"]}}

        transaction = [
            {"DeleteImage": cmd},
            {"DeleteVideo": cmd},
            {"DeleteBlob": cmd},
            {"DeleteClip": cmd},
            # DeleteDescriptorSet also deletes the descriptors
            {"DeleteDescriptorSet": cmd},

            # All the following should be deleted automatically once the
            # related objects are deleted.
            # We keep them here until ApertureDB fully implements this.
            {"DeleteBoundingBox": cmd},
            {"DeletePolygon": cmd},
            {"DeleteFrame": cmd},
            {"DeleteEntity": cmd},
            {"GetSchema": {"refresh": True}}
        ]

        try:
            response, _ = self.execute(transaction)
            schema = response[-1]["GetSchema"]
            if schema["status"] != 0:
                logger.error(f"status is non-zero: {response}")
            elif schema["connections"] is not None:
                logger.error(f"connections is not None: {response}")
            elif schema["entities"] is not None:
                logger.error(f"entities is not None: {response}")
            else:
                return True
        except BaseException as e:
            logger.exception(e)

        return False

    def user_log_message(self, message: str, level: str = "INFO") -> None:
        """
        Log a message to the user log.

        This is useful because it can later be seen in Grafana, not only as log entries in the AperturDB Logging dashboard, but also as event markers in the Aperture DB Status dahsboard.

        Args:
            message (str): The message to log.
            level (str): The level of the message. Default is "INFO".
        """
        assert level in ["INFO", "WARNING",
                         "ERROR"], f"Invalid log level: {level}"
        q = [{"UserLogMessage": {"text": message, "type": level}}]
        self.execute(q)



from typing import Set
import pandas as pd
import logging
from aperturedb.Subscriptable import Subscriptable
from dask import dataframe
import os
import multiprocessing as mp
import re


logger = logging.getLogger(__name__)

ENTITY_CLASS = "EntityClass"
CONSTRAINTS_PREFIX = "constraint_"
DATE_PREFIX = "date:"
PROPERTIES = "properties"
CONSTRAINTS = "constraints"

# This number is based on the partitions one wants to use per core.
PARTITIONS_PER_CORE = 10

# Use 90% os the CPU cores by default.
CORES_USED_FOR_PARALLELIZATION = 0.9


class CSVParser(Subscriptable):
    """
    **ApertureDB General CSV Parser for Loaders.**
    This operates in 2 modes:
    - **Normal Mode**: This is the default mode. It reads the CSV file into a Pandas DataFrame.
    - **Dask Mode**: This mode is used when the CSV file is too big to fit in memory, or multiprocessing is desired.
        It reads the CSV file into a Dask DataFrame.
        In Dask mode the CSV file is read in chunks, and the operations are performed on each chunk.
        The tricky bit is that the chunk size is not known till the loader is created, so the processing happens when ingest is called.
        So the Data CSV has another signature, where the df is passed explicitly.

    Typically, the response_handler is application specific, and loading does not break
    on errors in response_handlers, so the default behavior is to log the error and continue.
    If you want to break on errors, set strict_response_validation to True.
    """

    def __init__(self,
                 filename: str,
                 strict_response_validation: bool = False,
                 **kwargs):

        self.filename = filename
        self.strict_response_validation = strict_response_validation
        self.constraint_keyword = "if_not_found"

        # The following are extracted from the kwargs.
        self.blobs_relative_to_csv = "blobs_relative_to_csv" in kwargs and kwargs[
            "blobs_relative_to_csv"]
        self.use_dask = "use_dask" in kwargs and kwargs["use_dask"]
        df = kwargs["df"] if "df" in kwargs else None

        self.relative_path_prefix = os.path.dirname(self.filename) if self.blobs_relative_to_csv \
            else ""

        if not self.use_dask:
            if df is None:
                self.df = pd.read_csv(filename)
            else:
                self.df = df
        else:
            # It'll impact the number of partitions, and memory usage.
            # TODO: tune this for the best performance.
            cores_used = int(CORES_USED_FOR_PARALLELIZATION * mp.cpu_count())
            blocksize = os.path.getsize(
                self.filename) // (cores_used * PARTITIONS_PER_CORE)
            if blocksize == 0:
                cpus = mp.cpu_count()
                raise Exception(
                    f"CSV file too small to be read in parallel. Use normal mode. cpus: {cpus}")
            self.df = dataframe.read_csv(
                self.filename,
                blocksize=blocksize)

        # len for dask dataframe needs a client.
        if not self.use_dask and len(self.df) == 0:
            logger.error("Dataframe empty. Is the CSV file ok?")

        self.df = self.df.astype('object')
        self.header = list(self.df.columns.values)
        self.validate()

    def __len__(self):
        return len(self.df.index)

    def get_indexed_properties(self) -> Set[str]:
        if self.constraints_keys:
            return {self._parse_prop(k)[0] for k in self.constraints_keys}
        return set()

    def get_indices(self):
        raise NotImplementedError

    def _parse_prop(self, key, val=None):
        if key.startswith(CONSTRAINTS_PREFIX):
            key = key[len(CONSTRAINTS_PREFIX):]
        if key.startswith(DATE_PREFIX):
            key = key[len(DATE_PREFIX):]
            val = {"_date": val}
        return key, val

    def parse_properties(self, idx):

        properties = {}
        if len(self.props_keys) > 0:
            for key in self.props_keys:
                prop, value = self._parse_prop(key, self.df.loc[idx, key])
                if value == value:  # skips nan valies
                    properties[prop] = value
        return properties

    def parse_constraints(self, idx):

        constraints = {}
        if len(self.constraints_keys) > 0:
            for key in self.constraints_keys:
                prop, value = self._parse_prop(key, self.df.loc[idx, key])
                constraints[prop] = ["==", self.df.loc[idx, key]]
        return constraints

    def parse_other_constraint(self, constraint_name, keys, idx):

        other_constraints = {}
        if len(keys) > 0:
            for key in keys:
                res = re.search(f"^{constraint_name}_date(>|<)?:", key)
                if res is not None:
                    prop = key[len(res.group(0)):]  # remove prefix
                    sort = res.group(0)[-2:][:1]  # get character before :

                    if sort != ">" and sort != "<":
                        sort = "=="
                    other_constraints[prop] = [
                        sort, {"_date": self.df.loc[idx, key]}]
                else:
                    prop = key[len(constraint_name):]  # remove "prefix
                    op = "=="
                    if prop[0] in [">", "<", "!"]:
                        op = prop[0]
                        prop = str(prop[1:])

                    value = self.df.loc[idx, key]
                    other_constraints[prop] = [op, value]

        return other_constraints

    def _parsed_command(self, idx, custom_fields: dict = None, constraints: dict = None, properties: dict = None):
        if custom_fields == None:
            custom_fields = {}
        query = {
            self.command: custom_fields
        }
        if properties:
            query[self.command][PROPERTIES] = properties

        if constraints:
            query[self.command][self.constraint_keyword] = constraints

        return query

    def _basic_command(self, idx, custom_fields: dict = None):
        properties = self.parse_properties(idx)
        constraints = self.parse_constraints(idx)
        return self._parsed_command(idx, custom_fields, constraints, properties)

    def validate(self):

        Exception("Validation not implemented!")



import json
import os
from json import JSONEncoder
from pathlib import Path

import typer
from typing_extensions import Annotated

from aperturedb.cli.console import console
from aperturedb.Configuration import Configuration


class ObjEncoder(JSONEncoder):
    """
    A bit of boiler plate to allow us to serialize our Configuration object.
    """

    def default(self, o):
        return o.__dict__


APP_FULL_NAME = "CLI for aperturedb"
APP_NAME = "aperturedb"
APP_NAME_CLI = "adb"

app = typer.Typer()


def _config_file_path(as_global: bool):
    config_path: Path = Path(os.path.join(
        os.getcwd(),
        f".{APP_NAME}",
        f"{APP_NAME_CLI}.json"))
    if as_global:
        app_dir = typer.get_app_dir(APP_NAME)
        config_path: Path = Path(app_dir) / f"{APP_NAME_CLI}.json"
    return config_path


def check_configured(as_global: bool, show_error: bool = False):
    config_path = _config_file_path(as_global)
    file_exists = config_path.exists()
    if show_error:
        recommend = "Please run adb config create <name>"
        if not file_exists:
            console.log(f"{APP_NAME_CLI} is not configured yet. {recommend}")
    return file_exists


def get_configurations(file: str):
    configs = {}
    active = None
    with open(file) as config_file:
        configurations = json.loads(config_file.read())
        for c in filter(lambda key: key != "active", configurations):
            config = configurations[c]
            configs[c] = Configuration(
                name=c,
                host=config["host"],
                port=config["port"],
                username=config["username"],
                password=config["password"],
                use_rest=config["use_rest"],
                use_ssl=config["use_ssl"])
    active = configurations["active"]
    return configs, active


@app.command()
def ls(log_to_console: bool = True):
    """
    List the configurations with their details.
    """
    all_configs = {}
    for as_global in [True, False]:
        config_path = _config_file_path(as_global)
        context = "global" if as_global else "local"
        try:
            configs, active = get_configurations(config_path.as_posix())
            all_configs[context] = configs
            all_configs["active"] = active
        except FileNotFoundError:
            check_configured(as_global)
        except json.JSONDecodeError:
            check_configured(as_global)
            console.log("Failed to decode json")

    if "global" in all_configs or "local" in all_configs:
        if "global" in all_configs and len(all_configs["global"]) == 0 \
                and "local" in all_configs and len(all_configs["local"]) == 0:
            console.log(
                f"No configurations found. Please run adb config create <name>")
            return
        else:
            if log_to_console:
                console.log(f"Available configurations:")
                console.log(all_configs)

    else:
        console.log(all_configs)
        # Tried to parse global config as well as local, but failed.
        # Show user the error and bail.
        check_configured(as_global=True, show_error=True)

    return all_configs


@app.command()
def create(
        name: Annotated[str, typer.Argument(help="Name of this configuration for easy reference")],
        active: Annotated[bool, typer.Option(help="Set as active")] = False,
        as_global: Annotated[bool, typer.Option(
            help="Project level vs global level")] = True,
        host: Annotated[str, typer.Option(help="Host name")] = "localhost",
        port: Annotated[int, typer.Option(help="Port number")] = 55555,
        username: Annotated[str, typer.Option(help="Username")] = "admin",
        password: Annotated[str, typer.Option(help="Password")] = "admin",
        use_rest: Annotated[bool, typer.Option(help="Use REST")] = False,
        use_ssl: Annotated[bool, typer.Option(help="Use SSL")] = True,
        interactive: Annotated[bool, typer.Option(help="Interactive mode")] = True):
    """
    Create a new configuration for the client.
    """
    db_host = host
    db_port = port
    db_username = username
    db_password = password
    db_use_rest = use_rest
    db_use_ssl = use_ssl

    config_path = _config_file_path(as_global)
    configs = {}
    try:
        configs, ac = get_configurations(config_path.as_posix())
    except FileNotFoundError as e:
        config_path: Path = _config_file_path(as_global)
        parent_dir = os.path.dirname(config_path)
        if not os.path.exists(parent_dir):
            os.makedirs(parent_dir)
        active = True
    except json.JSONDecodeError:
        active = True

    if interactive:
        db_host = typer.prompt(f"Enter {APP_NAME} host name", default=db_host)
        db_port = typer.prompt(
            f"Enter {APP_NAME} port number", default=db_port)
        db_username = typer.prompt(
            f"Enter {APP_NAME} username", default=db_username)
        db_password = typer.prompt(
            f"Enter {APP_NAME} password", hide_input=True, default=db_password)
        db_use_rest = typer.confirm(f"Use REST", default=db_use_rest)
        db_use_ssl = typer.confirm(f"Use SSL", default=db_use_ssl)

    gen_config = Configuration(
        name=name,
        host=db_host,
        port=db_port,
        username=db_username,
        password=db_password,
        use_ssl=db_use_ssl,
        use_rest=db_use_rest
    )
    configs[name] = gen_config
    if active:
        configs["active"] = name
    else:
        configs["active"] = ac

    with open(config_path.as_posix(), "w") as config_file:
        config_file.write(json.dumps(configs, indent=2, cls=ObjEncoder))


@app.command()
def activate(
        name: Annotated[str, typer.Argument(help="Name of this configuration for easy reference")],
        as_global: Annotated[bool, typer.Option(help="Project level vs global level")] = True):
    """
    Set the default configuration.
    """
    global_config_path = _config_file_path(True)
    gc, ga = get_configurations(global_config_path)

    config_path = _config_file_path(as_global)
    configs = {}
    try:
        configs, ac = get_configurations(config_path.as_posix())
        if name not in configs and name not in gc:
            console.log(f"Configuration {name} not found")
            raise typer.Exit(code=2)
        configs["active"] = name
    except FileNotFoundError:
        check_configured(as_global=False) or \
            check_configured(as_global=True, show_error=True)
    except json.JSONDecodeError:
        check_configured(as_global=False) or \
            check_configured(as_global=True, show_error=True)

    with open(config_path.as_posix(), "w") as config_file:
        config_file.write(json.dumps(configs, indent=2, cls=ObjEncoder))




import logging
import numpy as np
from PIL import Image

logger = logging.getLogger(__name__)

error_message = """
CLIP transformer requires git+https://github.com/openai/CLIP.git and torch
Install with: pip install aperturedb[complete], followed by explicit install of CLIP.
Can be done with : "pip install git+https://github.com/openai/CLIP.git" in the same
venv as aperturedb.
"""

try:
    import clip
    import torch
    import cv2
except ImportError:
    logger.critical(error_message)
    exit(1)

descriptor_set = "ViT-B/16"
device = "cuda" if torch.cuda.is_available() else "cpu"
model, preprocess = clip.load(descriptor_set, device=device)


def generate_embedding(blob):
    global errors

    nparr = np.fromstring(blob, np.uint8)
    image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    image = preprocess(Image.fromarray(image)).unsqueeze(0).to(device)

    image_features = model.encode_image(image)
    embedding = None
    if device == "cuda":
        image_features = image_features.float()
        embedding = image_features.detach().cpu().numpy().tobytes()
    else:
        embedding = image_features.detach().numpy().tobytes()

    return embedding



from enum import Enum
import logging
from typing import List, Optional
import typer
import os
import time
import numpy as np
import pandas as pd

from typing_extensions import Annotated
from aperturedb.ParallelLoader import ParallelLoader
from aperturedb.ImageDataCSV import ImageDataCSV
from aperturedb.BBoxDataCSV import BBoxDataCSV
from aperturedb.EntityDataCSV import EntityDataCSV
from aperturedb.BlobDataCSV import BlobDataCSV
from aperturedb.ConnectionDataCSV import ConnectionDataCSV
from aperturedb.PolygonDataCSV import PolygonDataCSV
from aperturedb.VideoDataCSV import VideoDataCSV
from aperturedb.DescriptorDataCSV import DescriptorDataCSV
from aperturedb.DescriptorSetDataCSV import DescriptorSetDataCSV

from aperturedb.Utils import create_connector
from aperturedb.Query import ObjectType
from aperturedb.cli.console import console

from aperturedb.Utils import import_module_by_path

from tqdm import tqdm

logger = logging.getLogger(__file__)
app = typer.Typer()


IngestType = Enum('IngestType', {k: str(k) for k in ObjectType._member_names_})

# This list needs to be updated when new transformers are added
# This enables CLI params.


class TransformerType(str, Enum):
    common_properties = "common_properties"
    image_properties = "image_properties"
    clip_pytorch_embeddings = "clip_pytorch_embeddings"
    facenet_pytorch_embeddings = "facenet_pytorch_embeddings"


def _debug_samples(data, sample_count, module_name):
    import json
    print(f"len(data)={len(data)}")
    for i, r in enumerate(data[0:sample_count]):
        query, blobs = r
        with open(module_name + f"_{i}" + ".raw", "w") as f:
            f.write(str(query))
        with open(module_name + f"_{i}" + ".json", "w") as f:
            f.write(json.dumps(query, indent=2))
        for j, blob in enumerate(blobs):
            with open(module_name + f"_{i}_{j}" + ".jpg", "wb") as f:
                f.write(blob)


def _apply_pipeline(data, transformers: List[str], **kwargs):
    pipeline = _create_pipeline(transformers)
    console.log("Applying Pipeline: \r\n" +
                "\r\n=>".join([f"{stage.__name__}[{kwargs}]" for stage in pipeline]))
    for transformer in pipeline:
        data = transformer(data, **kwargs)
        data.sample_count = data.data.sample_count
    return data


def _create_pipeline(transformers: List[str]):
    from aperturedb.transformers.common_properties import CommonProperties
    from aperturedb.transformers.image_properties import ImageProperties
    from aperturedb.transformers.clip_pytorch_embeddings import CLIPPyTorchEmbeddings
    from aperturedb.transformers.facenet_pytorch_embeddings import FacenetPyTorchEmbeddings

    # Actual collection of transformers, packaged with aperturedb.
    built_in_transformers = {
        "common_properties": CommonProperties,
        "image_properties": ImageProperties,
        "clip_pytorch_embeddings": CLIPPyTorchEmbeddings,
        "facenet_pytorch_embeddings": FacenetPyTorchEmbeddings
    }
    pipeline = []
    for transformer_name in transformers:
        transformer = None
        if transformer_name in built_in_transformers:
            transformer = built_in_transformers[transformer_name]
        else:
            try:
                transformer_module = import_module_by_path(transformer_name)
                transformer = getattr(transformer_module,
                                      transformer_module.__name__)
            except Exception as e:
                console.log(
                    f"Could not load transformer {transformer_name}: {e}")
                exit(1)
        pipeline.append(transformer)
    return pipeline


@app.command()
def from_generator(filepath: Annotated[str, typer.Argument(
    help="Path to python module for ingestion [BETA]")],
    sample_count: Annotated[int, typer.Option(
        help="Number of samples to ingest (-1 for all)")] = -1,
    debug: Annotated[bool, typer.Option(
        help="Debug mode")] = False,
    batchsize: Annotated[int, typer.Option(
        help="Size of the batch")] = 1,
    num_workers: Annotated[int, typer.Option(
        help="Number of workers for ingestion")] = 1,
    transformer: Annotated[Optional[List[TransformerType]], typer.Option(
        help="Apply transformer to the pipeline [Can be specified multiple times]")] = None,
    user_transformer: Annotated[Optional[List[str]], typer.Option(
        help="Apply user transformer to the pipeline as path to file [Can be specified multiple times]")] = None,
):
    """
    Ingest data from a Data generator [BETA].
    """
    db = create_connector()
    loader = ParallelLoader(db)

    module = import_module_by_path(filepath)

    # The assumption is that the class name is same as that of the module name
    module_name  = os.path.basename(filepath)[:-3]
    mclass = getattr(module, module_name)

    # This is the dynamically loaded data generator.
    # tested with CelebADataKaggle.py, CocoDataPytorch.py and Cifar10DataTensorflow.py in examples
    start = time.time()
    data = mclass()
    data.sample_count = len(data) if sample_count == -1 else sample_count
    console.log(f"Data generator loaded in {time.time() - start} seconds")

    if transformer or user_transformer:
        transformer = transformer or []
        user_transformer = user_transformer or []
        all_transformers = transformer + user_transformer
        data = _apply_pipeline(data, all_transformers,
                               adb_data_source=f"{module_name}.{mclass.__name__}")
    else:
        console.log("No transformer applied")

    if debug:
        _debug_samples(data, sample_count, module_name)
    else:
        loader.ingest(
            data,
            stats=True,
            batchsize=batchsize,
            numthreads=num_workers)

    while hasattr(data, "ncalls"):
        console.log(
            f"Calls to {data}.getitem = {data.ncalls} time={data.cumulative_time}")
        data = data.data


@app.command()
def from_csv(filepath: Annotated[str, typer.Argument(
    help="Path to csv for ingestion")],
    batchsize: Annotated[int, typer.Option(
        help="Size of the batch")] = 1,
    num_workers: Annotated[int, typer.Option(
        help="Number of workers for ingestion")] = 1,
    stats: Annotated[bool, typer.Option(
        help="Show realtime statistics with summary")] = True,
    use_dask: Annotated[bool, typer.Option(
        help="Use dask based parallelization")] = False,
    ingest_type: Annotated[IngestType, typer.Option(
        help="Parser for CSV file to be used")] = IngestType.IMAGE,
    blobs_relative_to_csv: Annotated[bool, typer.Option(
        help="If true, the blob path is relative to the CSV file")] = True,
    transformer: Annotated[Optional[List[TransformerType]], typer.Option(
        help="Apply transformer to the pipeline [Can be specified multiple times]")] = None,
    user_transformer: Annotated[Optional[List[str]], typer.Option(
        help="Apply user transformer to the pipeline as path to file [Can be specified multiple times.]")] = None,
    sample_count: Annotated[int, typer.Option(
        help="Number of samples to ingest (-1 for all)")] = -1,
    debug: Annotated[bool, typer.Option(
        help="Debug mode")] = False,
):
    """
    Ingest data from a pre generated CSV file.
    """

    ingest_types = {
        IngestType.BLOB: BlobDataCSV,
        IngestType.BOUNDING_BOX: BBoxDataCSV,
        IngestType.CONNECTION: ConnectionDataCSV,
        IngestType.DESCRIPTOR: DescriptorDataCSV,
        IngestType.DESCRIPTORSET: DescriptorSetDataCSV,
        IngestType.ENTITY: EntityDataCSV,
        IngestType.IMAGE: ImageDataCSV,
        IngestType.POLYGON: PolygonDataCSV,
        IngestType.VIDEO: VideoDataCSV
    }

    data = ingest_types[ingest_type](filepath, use_dask=use_dask,
                                     blobs_relative_to_csv=blobs_relative_to_csv)
    data.sample_count = len(data) if sample_count == -1 else sample_count
    if transformer or user_transformer:
        transformer = transformer or []
        user_transformer = user_transformer or []
        all_transformers = transformer + user_transformer
        data = _apply_pipeline(data, all_transformers,
                               adb_data_source=f"{ingest_type}.{os.path.basename(filepath)}")
    else:
        console.log("No transformer applied")
    db = create_connector()
    console.log(db)

    loader = ParallelLoader(db)
    if debug:
        _debug_samples(data, sample_count, filepath)
    else:
        loader.ingest(data, batchsize=batchsize,
                      numthreads=num_workers, stats=stats)


@app.command()
def generate_embedding_csv_from_image_csv(
    input_file: Annotated[str, typer.Argument(
        help="Path to csv for ingestion")],
    set_name: Annotated[str, typer.Argument(
        help="Descriptor set name to be associated with the embeddings")],
    transformer: Annotated[Optional[List[TransformerType]], typer.Option(
        help="Apply transformer to the pipeline [Can be specified multiple times]")] = None,
    user_transformer: Annotated[Optional[List[str]], typer.Option(
        help="Apply user transformer to the pipeline as path to file [Can be specified multiple times.]")] = None,
    sample_count: Annotated[int, typer.Option(
        help="Number of samples to ingest (-1 for all)")] = -1,
):
    """
    Create a DescriptorDataCSV from an ImageDataCSV, using the specified transformer.
    Also save embeddings in a separate file, and generate the ConnectionDataCSV
    for linking the images to the descriptors.
    The ingestion of the embeddings and the connection data is not done here.
    image_properties transformer must be used at actual ingestion.
    It is also applied by default for CSV generation.
    """
    data = ImageDataCSV(input_file)
    data.sample_count = len(data) if sample_count == -1 else sample_count
    if transformer or user_transformer:
        transformer = transformer or []
        user_transformer = user_transformer or []
        all_transformers = transformer + user_transformer
        if len(all_transformers) == 0:
            console.log(
                "No transformer specified . Generating embeddings from raw images requires a transformer.")
            typer.Abort()
        all_transformers.insert(0, "image_properties")
        data = _apply_pipeline(data, all_transformers,
                               adb_data_source=f"{os.path.basename(input_file)}")
    filename = f"{os.path.basename(input_file)}_{all_transformers[-1]}"
    metadata = []
    connection = []
    embeddings = []

    for i in tqdm(range(data.sample_count)):
        d = data[i]
        embeddings.append(np.frombuffer(d[1][1], dtype=np.float32))
        descriptor_id = f"{set_name}_{os.path.basename(input_file)}_{i}"
        metadata.append({
            "filename": f"{filename}.npy",
            "index": i,
            "set": set_name,
            "id": descriptor_id,
            "constraint_id": descriptor_id
        })
        connection.append({
            "ConnectionClass": "image_descriptor",
            "_Image@adb_image_sha256": d[0][0]["AddImage"]["properties"]["adb_image_sha256"],
            "_Descriptor@id": descriptor_id,
        })

    with open(f"{filename}.npy", "wb") as f:
        np.save(f, embeddings)
    pd.json_normalize(metadata).to_csv(
        f"{filename}_metadata.adb.csv", index=False)
    pd.json_normalize(connection).to_csv(
        f"{filename}_connection.adb.csv", index=False)



import os
import time


class ProgressBar(object):

    def __init__(self, filename="", use_last=0):

        if filename != "":
            self.fd = open(filename, 'w')
            self.file = True
            self.filename = filename
        else:
            self.file = False

        self.use_last = 50 if use_last == 0 else use_last
        self.progress = 0.0
        self.eta = 0

        self.progress_arr  = []
        self.progress_time = []

        self.done = False

        self.last_msg_call = time.time()

    def __del__(self):

        if self.file:
            self.fd.close()

    def msg(self):

        self.progress = 1 if self.progress > 1 else self.progress

        progress = "{:.2f}".format(self.progress * 100)
        eta = "{:.2f}".format(self.eta)
        msg = "Progress: " + progress
        msg += "% - ETA(" + self.eta_unit + "): " + eta

        if self.done:
            msg += "\n"

        return msg

    def updatefile(self):

        self.fd.write("\r" + self.msg())
        self.fd.flush()

    def updatescreen(self):

        print('\r', end="")
        print(self.msg(), end="", flush=True)

    def compute_stats(self, progress):

        if int(progress) == 1:
            self.done = True

        if progress < self.progress:
            self.progress = []
            self.progress_time = []

        self.progress = progress
        self.progress_arr.append(progress)
        self.progress_time.append(time.time())

        progress_per_sec = 0

        firsttime = True
        for (progress, rec_time) in zip(self.progress_arr, self.progress_time):

            if firsttime:
                firsttime = False
            else:
                delta = rec_time - prev_time
                progress_diff = progress - prev_progress
                progress_per_sec += progress_diff / delta

            prev_progress = progress
            prev_time     = rec_time

        samples = len(self.progress_arr)
        avg_progress_per_sec = 0

        if samples >= 2:
            avg_progress_per_sec = progress_per_sec / (samples - 1)

        now = time.time()
        # If we called this function less than a second ago,
        # we don't update ETE because jitters too much.
        if now - self.last_msg_call < 1 and self.eta > 0:
            return

        self.last_msg_call = now

        if avg_progress_per_sec > 0:
            self.eta = (1 - progress) / avg_progress_per_sec
        else:
            self.eta = 0

        if self.eta > 3600.0:
            self.eta = self.eta / 3600.0
            self.eta_unit = "h"
        elif self.eta > 60.0:
            self.eta = self.eta / 60.0
            self.eta_unit = "m"
        else:
            self.eta_unit = "s"

        if len(self.progress_arr) > self.use_last:
            self.progress_arr  = self.progress_arr[-(self.use_last):]
            self.progress_time = self.progress_time[-(self.use_last):]

    def update(self, progress):

        if self.done:
            return

        self.compute_stats(progress)

        if self.file:
            self.updatefile()
        else:
            self.updatescreen()



import time
import requests
import boto3

import numpy as np
import cv2

from aperturedb import CSVParser
from aperturedb.EntityUpdateDataCSV import SingleEntityUpdateDataCSV
from aperturedb.BlobNewestDataCSV import BlobNewestDataCSV
from aperturedb.SparseAddingDataCSV import SparseAddingDataCSV
from aperturedb.Sources import Sources
import logging
import os

logger = logging.getLogger(__name__)

HEADER_PATH = "filename"
HEADER_URL = "url"
HEADER_S3_URL = "s3_url"
HEADER_GS_URL = "gs_url"
PROPERTIES = "properties"
CONSTRAINTS = "constraints"
IMG_FORMAT = "format"


class ImageDataProcessor():
    """
    **Processing for Image data, used when loading images**
    """

    def __init__(self, check_image, n_download_retries):
        self.loaders = [self.load_image, self.load_url,
                        self.load_s3_url, self.load_gs_url]
        self.source_types = [HEADER_PATH,
                             HEADER_URL, HEADER_S3_URL, HEADER_GS_URL]

        self.check_image = check_image
        self.n_download_retries = n_download_retries

    def set_processor(self, use_dask: bool, source_type):
        self.source_type = source_type
        if use_dask == False and self.source_type == HEADER_S3_URL:
            # The connections by boto3 cause ResourceWarning. Known
            # issue: https://github.com/boto/boto3/issues/454
            s3_client = boto3.client('s3')
            self.sources = Sources(
                self.n_download_retries,
                s3_client=s3_client)
        else:
            self.sources = Sources(self.n_download_retries)

    def get_indices(self):
        return {
            "entity": {
                "_Image": self.get_indexed_properties()
            }
        }

    def load_image(self, filename):

        if self.check_image:
            try:
                a = cv2.imread(filename)
                if a.size <= 0:
                    logger.error(f"IMAGE SIZE ERROR: {filename}")
                    return False, None
            except Exception as e:
                logger.error(f"IMAGE ERROR: {filename}")
                logger.exception(e)

        return self.sources.load_from_file(filename)

    def check_image_buffer(self, img):

        if not self.check_image:
            return True

        try:
            # This is a full decoding of the image, a compute intensive operation.
            # decoded_img = cv2.imdecode(img, cv2.IMREAD_COLOR)
            # if decoded_img is None:
            #     logger.error("Error decoding image")
            #     print(img[0], img[1], img[2], img[3], img[4], img[5], img[6], img[7])
            #     return False

            # Instead, we do a much lighter check on the header of the image,
            # which is already in memory.
            # Check header signature for file format:
            # https://en.wikipedia.org/wiki/List_of_file_signatures

            # NOTE: Make sure we are checking for the same values
            #       as Image.cc on ApertureDB
            # Ideally, we don't want this logic here because it is replicated
            # on the server side, and we have to maintain it in both places.
            # But because it is expensive to send the image to the server
            # only to find out it is bad, we do it here.

            # JPEG
            if img[0] == 255 and img[1] == 216:
                return True
            # PNG
            if img[0] == 137 and img[1] == 80 and img[2] == 78 and img[3] == 71:
                return True
            # TIF
            if img[0] == 73 and img[1] == 73 and img[2] == 42:
                return True

        except Exception as e:
            logger.exception(e)
            return False

        return False  # When header is not recognized

    def load_url(self, url):
        return self.sources.load_from_http_url(url, self.check_image_buffer)

    def load_s3_url(self, s3_url):
        return self.sources.load_from_s3_url(s3_url, self.check_image_buffer)

    def load_gs_url(self, gs_url):
        return self.sources.load_from_gs_url(gs_url, self.check_image_buffer)


class ImageDataCSV(CSVParser.CSVParser, ImageDataProcessor):
    """**ApertureDB Image Data.**

    This class loads the Image Data which is present in a CSV file,
    and converts it into a series of ApertureDB queries.


    :::note Is backed by a CSV file with the following columns:

    ``filename``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

    OR

    ``url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

    OR

    ``s3_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

    OR

    ``gs_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``
    :::

    Example CSV file::

        filename,id,label,constraint_id,format
        /home/user/file1.jpg,321423532,dog,321423532,jpg
        /home/user/file2.jpg,42342522,cat,42342522,png
        ...

    Example usage:

    ``` python

        data = ImageDataCSV("/path/to/ImageData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    ```

    or

    ``` bash

        adb ingest from-csv /path/to/ImageData.csv --ingest-type IMAGE
    ```


    :::info
    In the above example, the constraint_id ensures that an Image with the specified
    id would be only inserted if it does not already exist in the database.
    :::
    """

    def __init__(self, filename: str, check_image: bool = True, n_download_retries: int = 3, **kwargs):

        ImageDataProcessor.__init__(
            self, check_image, n_download_retries)
        CSVParser.CSVParser.__init__(self, filename, **kwargs)

        source_type = self.header[0]
        self.set_processor(self.use_dask, source_type)

        self.format_given = IMG_FORMAT in self.header
        self.props_keys = [x for x in self.header[1:]
                           if not x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
        self.props_keys = [
            x for x in self.props_keys if x != IMG_FORMAT]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONSTRAINTS_PREFIX)]

        if self.source_type not in self.source_types:
            logger.error("Source not recognized: " + self.source_type)
            raise Exception("Error loading image: " + filename)
        self.source_loader = {
            st: sl for st, sl in zip(self.source_types, self.loaders)
        }

        self.relative_path_prefix = os.path.dirname(self.filename) \
            if self.source_type == HEADER_PATH and self.blobs_relative_to_csv else ""

        self.command = "AddImage"

    def getitem(self, idx):
        idx = self.df.index.start + idx

        image_path = os.path.join(
            self.relative_path_prefix, self.df.loc[idx, self.source_type])
        img_ok, img = self.source_loader[self.source_type](image_path)

        if not img_ok:
            logger.error("Error loading image: " + image_path)
            raise Exception("Error loading image: " + image_path)

        q = []
        blobs = []
        custom_fields = {}
        if self.format_given:
            custom_fields["format"] = self.df.loc[idx, IMG_FORMAT]
        ai = self._basic_command(idx, custom_fields)
        # Each getitem query should be properly defined with a ref.
        # A ref shouldb be added to each of the commands from getitem implementation.
        # This is because a transformer or ref updater in the PQ
        # will need to know which command to update.
        ai[self.command]["_ref"] = 1
        blobs.append(img)
        q.append(ai)

        return q, blobs

    def get_indices(self):
        return {
            "entity": {
                "_Image": self.get_indexed_properties()
            }
        }

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] not in self.source_types:
            field = self.header[0]
            allowed = ", ".join(self.source_types)
            raise Exception(
                f"Error with CSV file field: {field}. Allowed values: {allowed}")


class ImageUpdateDataCSV(SingleEntityUpdateDataCSV, ImageDataProcessor):
    """
    **ApertureDB Image CSV Parser for Adding an Image and updating the properties on the image.**
    Usage is in EntityUpdateDataCSV.

    Note that this class will not change a blob in an existing entity. If looking to change a blob in
     an existing entity, look at ImageForceNewestDataCSV, but be careful of all caveats.

    """

    def __init__(self, filename: str, check_image: bool = True, n_download_retries: int = 3, **kwargs):
        ImageDataProcessor.__init__(
            self, check_image, n_download_retries)
        SingleEntityUpdateDataCSV.__init__(
            self, "Image", filename, **kwargs)

        source_type = self.header[0]
        self.set_processor(self.use_dask, source_type)

        # this class loads a blob, so must set the first query to have a blob.
        self.blobs_per_query = [1, 0]
        self.format_given = IMG_FORMAT in self.header
        self.props_keys = list(filter(lambda prop: prop not in [
                               IMG_FORMAT, "filename"], self.props_keys))
        if self.source_type not in self.source_types:
            logger.error("Source not recognized: " + self.source_type)
            raise Exception("Error loading image: " + filename)
        self.source_loader = {
            st: sl for st, sl in zip(self.source_types, self.loaders)
        }

    def getitem(self, idx):
        blob_set = []
        [query_set, empty_blobs] = super().getitem(idx)

        image_path = os.path.join(
            self.relative_path_prefix, self.df.loc[idx, self.source_type])
        img_ok, img = self.source_loader[self.source_type](image_path)
        if not img_ok:
            logger.error("Error loading image: " + image_path)
            raise Exception("Error loading image: " + image_path)
        # element has 2 queries, each with 1 blob.
        blob_set = [[img], []]
        # must wrap the blob return for this item in a list
        return [query_set, [blob_set]]

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] not in self.source_types:
            raise Exception(
                f"Error with CSV file field: {self.header[0]}. Must be first field")
        else:
            SingleEntityUpdateDataCSV.validate(self)


class ImageForceNewestDataCSV(BlobNewestDataCSV, ImageDataProcessor):
    """
    **ApertureDB Image CSV Parser for Maintaining a Blob set with changing blob data.**
    See BlobNewestDataCSV for usage.
    """

    def __init__(self, filename, check_image=True, n_download_retries=3, **kwargs):
        ImageDataProcessor.__init__(
            self, check_image, n_download_retries)
        BlobNewestDataCSV.__init__(
            self, "Image", filename, **kwargs)

        source_type = self.header[0]
        self.set_processor(self.use_dask, source_type)

        self.format_given = IMG_FORMAT in self.header
        self.props_keys = list(filter(lambda prop: prop not in [
                               IMG_FORMAT, "filename"], self.props_keys))
        if self.source_type not in self.source_types:
            logger.error("Source not recognized: " + self.source_type)
            raise Exception("Error loading image: " + filename)
        self.source_loader = {
            st: sl for st, sl in zip(self.source_types, self.loaders)
        }

    def read_blob(self, idx):
        relative_path_prefix = os.path.dirname(self.filename) \
            if self.source_type == HEADER_PATH else ""
        image_path = os.path.join(
            relative_path_prefix, self.df.loc[idx, self.source_type])
        img_ok, img = self.source_loader[self.source_type](image_path)
        if not img_ok:
            logger.error("Error loading image: " + image_path)
            raise Exception("Error loading image: " + image_path)
        return img


class ImageSparseAddDataCSV(SparseAddingDataCSV, ImageDataProcessor):
    """
    **ApertureDB Spare Loading Image Data.**

    ImageSparseAddDataCSV should be used the same as ImageDataCSV.

    See SparseAddingDataCSV for description of when to use this versus ImageDataCSV.


    Example CSV file::

        filename,id,label,constraint_id,format
        /home/user/file1.jpg,321423532,dog,321423532,jpg
        /home/user/file2.jpg,42342522,cat,42342522,png
        ...

    Example usage:

    ```python

        data = ImageSparseAddDataCSV("/path/to/ImageData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    ```
    """

    def __init__(self, filename: str, check_image: bool = True, n_download_retries: int = 3, **kwargs):
        ImageDataProcessor.__init__(
            self, check_image, n_download_retries)
        SparseAddingDataCSV.__init__(self, "Image", filename, **kwargs)
        source_type = self.header[0]
        self.set_processor(self.use_dask, source_type)

        self.format_given = IMG_FORMAT in self.header
        self.props_keys = list(filter(lambda prop: prop not in [
                               IMG_FORMAT, "filename"], self.props_keys))
        if self.source_type not in self.source_types:
            logger.error("Source not recognized: " + self.source_type)
            raise Exception("Error loading image: " + filename)
        self.source_loader = {
            st: sl for st, sl in zip(self.source_types, self.loaders)
        }

    def getitem(self, idx):
        blob_set = []
        [query_set, empty_blobs] = super().getitem(idx)

        image_path = os.path.join(
            self.relative_path_prefix, self.df.loc[idx, self.source_type])
        img_ok, img = self.source_loader[self.source_type](image_path)

        if not img_ok:
            logger.error("Error loading image: " + image_path)
            raise Exception("Error loading image: " + image_path)
        # element has 2 queries, only second has blob
        blob_set = [[], [img]]
        # must wrap the blob return for this item in a list
        return [query_set, [blob_set]]



from enum import Enum
from typing import Annotated

import typer

from aperturedb.Utils import Utils, create_connector
from aperturedb.cli.console import console

app = typer.Typer()


class CommandTypes(str, Enum):
    STATUS = "status"
    SUMMARY = "summary"
    REMOVE_ALL = "remove_all"
    REMOVE_INDEXES = "remove_indexes"


def confirm(command: CommandTypes, force: bool):
    if force:
        return True
    console.print("Danger", style="bold red")
    console.log(f"This will execute {command}.")
    response = typer.prompt("Are you sure you want to continue? [y/N]")
    if response.lower() != "y":
        typer.echo("Aborting...")
        raise typer.Abort()
    return True


@app.command()
def execute(command: CommandTypes,
            force: Annotated[bool, typer.Option(help="Do not confirm")] = False):
    utils = Utils(create_connector())
    available_commands = {
        CommandTypes.STATUS: lambda: print(utils),
        CommandTypes.SUMMARY: utils.summary,
        CommandTypes.REMOVE_ALL: lambda: confirm(
            CommandTypes.REMOVE_ALL, force) and utils.remove_all_objects(),
        CommandTypes.REMOVE_INDEXES: lambda: confirm(
            CommandTypes.REMOVE_INDEXES, force) and utils.remove_all_indexes(),
    }

    available_commands[command]()


class LogLevel(str, Enum):
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"


@app.command()
def log(
    message: Annotated[str, typer.Argument(help="The message to log")],
    level: LogLevel = LogLevel.INFO
):
    """
    Log a message to the user log.

    This is useful because it can later be seen in Grafana, not only as log entries in the AperturDB Logging dashboard, but also as event markers in the Aperture DB Status dahsboard.
    """
    utils = Utils(create_connector())
    utils.user_log_message(message, level=level.value)


@app.command()
def visualize_schema(
    filename: str = "schema",
    format: str = "png"
):
    """
    Visualize the schema of the database.

    This will create a file with the schema of the database in the specified format.

    Relies on graphviz to be installed.
    """

    utils = Utils(create_connector())
    s = utils.visualize_schema()
    result = s.render(filename, format=format)
    print(result)



from __future__ import annotations
from typing import Any

from aperturedb.Entities import Entities
from IPython.display import HTML, display
from aperturedb.NotebookHelpers import display_annotated_video
from ipywidgets import widgets


class Videos(Entities):
    """
    **The object mapper representation of videos in ApertureDB.**

    This class is a layer on top of the native query.
    It facilitates interactions with videos in the database in the pythonic way.
    """
    db_object = "_Video"

    def getitem(self, idx):
        item = super().getitem(idx)
        if self.blobs:
            if 'preview' not in item:
                item['preview'] = self.get_blob(item)
        return item

    def inspect(self, show_preview: bool = True, meta = None) -> Any:
        if meta == None:
            def meta(x): return []
        df = super().inspect()
        if show_preview == True:
            op = widgets.Output()
            with op:
                df['preview'] = df.apply(lambda x: display_annotated_video(
                    x["preview"], bboxes=meta(x)), axis=1)
                display(HTML(
                    "<div style='max-width: 100%; overflow: auto;'>" +
                    df.to_html(escape=False)
                    + "</div>"
                ))
            return op
        else:
            return df



import pytest
from aperturedb.Connector import Connector
import dbinfo

import logging
logger = logging.getLogger(__name__)


class TestBadResponses():

    def test_AuthFailure(self, monkeypatch):

        def failed_auth_query(conn_obj, ignored_query):
            # generate a response from the server which is not the expected Auth result.
            # _query returns the server response json and an array of blobs.
            return ({"info": "Internal Server Error 42", "status": -1, "ignored": ignored_query}, [])

        monkeypatch.setattr(Connector, "_query", failed_auth_query)

        with pytest.raises(Exception) as conn_exception:
            db = Connector(
                host = dbinfo.DB_TCP_HOST,
                port = dbinfo.DB_TCP_PORT,
                user = dbinfo.DB_USER,
                password = dbinfo.DB_PASSWORD,
                use_ssl = True)
            db.query([{
                "FindImage": {
                    "results": {
                        "limit": 5
                    }
                }
            }])

        assert "Unexpected response" in str(conn_exception.value)



#
# The MIT License
#
# @copyright Copyright (c) 2017 Intel Corporation
# @copyright Copyright (c) 2021 ApertureData Inc
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify,
# merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
from __future__ import annotations
import os
import requests
import time
import json
import logging

from threading import Lock
from types import SimpleNamespace
from dataclasses import dataclass
from aperturedb.Connector import Connector
from aperturedb.Configuration import Configuration

logger = logging.getLogger(__name__)

PROTOCOL_VERSION = 1


class ConnectorRest(Connector):
    """
    **Class to use ApertureDB's REST interface**

    Args:
        str (host): Address of the host to connect to.
        int (port): Port to connect to.
        str (user): Username to specify while establishing a connection.
        str (password): Password to specify while connecting to the db.
        str (token): Token to use while connecting to the database.
        bool (use_ssl): Use SSL to encrypt communication with the database.
    """

    def __init__(self, host="localhost", port=None,
                 user="", password="", token="",
                 use_ssl=True, shared_data=None,
                 config: Configuration = None):
        self.use_keepalive = False
        super().__init__(
            host=host,
            port=port,
            user=user,
            password=password,
            token=token,
            use_ssl=use_ssl,
            shared_data=shared_data,
            config=config)

        if port is None:
            self.port = 443 if use_ssl else 80
        else:
            self.port = port

        self.connected = False
        # Session is useful because it does not add "Connection: close header"
        # Since we will be making same call to the same URL, making a session
        # REF: https://requests.readthedocs.io/en/latest/user/advanced/
        self.http_session = requests.Session()

        self.last_response   = ''
        self.last_query_time = 0

        self.url = ('https' if self.use_ssl else 'http') + \
            '://' + host + ':' + str(port) + '/api/'

        self.token = token

    def __del__(self):
        logger.info("Done with connector REST.")
        self.http_session.close()

    def _query(self, query, blob_array = [], try_resume=True):
        response_blob_array = []
        # Check the query type
        if not isinstance(query, str):  # assumes json
            query_str = json.dumps(query)
        else:
            query_str = query

        files = [
            ('query', (None, query_str)),
        ]

        for blob in blob_array:
            files.append(('blobs', blob))

        # Set Auth token, only when not authenticated before
        if self.shared_data.session and self.shared_data.session.valid():
            headers = {'Authorization': "Bearer " +
                       self.shared_data.session.session_token}
        else:
            headers = None
        tries = 0
        response = SimpleNamespace()
        response.status_code = 0
        while tries < self.config.retry_max_attempts:
            tries += 1
            try:
                response = self.http_session.post(self.url,
                                                  headers = headers,
                                                  files   = files,
                                                  verify  = self.use_ssl)
                if response.status_code == 200:
                    # Parse response:
                    json_response       = json.loads(response.text)
                    import base64
                    response_blob_array = [base64.b64decode(
                        b) for b in json_response['blobs']]
                    self.last_response  = json_response["json"]
                    break
                logger.error(
                    f"Response not OK = {response.status_code} {response.text[:1000]}\n\
                        attempt [{tries}/3] .. PID = {os.getpid()}")
            except Exception as e:
                logger.error(f"Exception during http.post = {e}\n\
                        attempt [{tries}/3] .. PID = {os.getpid()}")

            time.sleep(self.config.retry_interval_seconds)

            self.connect()
            if try_resume:
                self._renew_session()

        if tries == self.config.retry_max_attempts:
            raise Exception(
                f"Could not query ApertureDB {self.config} using REST.")
        return (self.last_response, response_blob_array)

    def _connect(self):
        logger.info("Connecting to ApertureDB using REST")
        self.connected = True



import time
import os
import logging
from typing import Union

import torch
import torch.distributed as dist
from aperturedb import Images
from aperturedb import PyTorchDataset
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Dataset

from aperturedb.ConnectorRest import ConnectorRest

logger = logging.getLogger(__name__)


class TestTorchDatasets():
    def validate_dataset(self, dataset: Union[DataLoader, Dataset], expected_length):
        start = time.time()

        count = 0
        # Iterate over dataset.
        for img in dataset:
            if len(img[0]) < 0:
                logger.error("Empty image?")
                assert True == False
            count += len(img[1]) if isinstance(dataset, DataLoader) else 1
        assert count == expected_length

        time_taken = time.time() - start
        if time_taken != 0:
            logger.info(f"Throughput (imgs/s): {len(dataset) / time_taken}")

    def test_nativeContraints(self, db, utils, images):
        assert len(images) > 0
        # This is a hack against a bug in batch API.
        dim = 224 if isinstance(db, ConnectorRest) else 225
        query = [{
            "FindImage": {
                "constraints": {
                    "age": [">=", 0]
                },
                "operations": [
                    {
                        "type": "resize",
                        "width": dim,
                        "height": dim
                    }
                ],
                "results": {
                    "list": ["license"]
                }
            }
        }]

        dataset = PyTorchDataset.ApertureDBDataset(
            db, query, label_prop="license")

        self.validate_dataset(dataset, utils.count_images())

    def test_datasetWithMultiprocessing(self, db, utils, images):
        len_limit = utils.count_images()
        # This is a hack against a bug in batch API.
        # TODO Fixme
        dim = 224 if isinstance(db, ConnectorRest) else 225
        query = [{
            "FindImage": {
                "constraints": {
                    "age": [">=", 0]
                },
                "operations": [
                    {
                        "type": "resize",
                        "width": dim,
                        "height": dim
                    }
                ],
                "results": {
                    "list": ["license"],
                    "limit": len_limit
                }
            }
        }]

        dataset = PyTorchDataset.ApertureDBDataset(
            db, query, label_prop="license")

        self.validate_dataset(dataset, len_limit)

        # Distributed Data Loader Setup

        # Needed for init_process_group
        os.environ['MASTER_ADDR'] = 'localhost'
        os.environ['MASTER_PORT'] = '12355'

        dist.init_process_group("gloo", rank=0, world_size=1)

        # === Distributed Data Loader Sequential
        batch_size = 10
        data_loader = DataLoader(
            dataset,
            batch_size=batch_size,          # pick random values here to test
            num_workers=4,          # num_workers > 1 to test multiprocessing works
            pin_memory=True,
            drop_last=True,
        )

        self.validate_dataset(data_loader, len_limit)
        # === Distributed Data Loader Shuffler

        # This will generate a random sampler, which will make the use
        # of batching wasteful
        sampler     = torch.utils.data.DistributedSampler(
            dataset, shuffle=True)

        data_loader = DataLoader(
            dataset,
            sampler=sampler,
            batch_size=batch_size,          # pick random values here to test
            num_workers=4,          # num_workers > 1 to test multiprocessing works
            pin_memory=True,
            drop_last=True,
        )

        self.validate_dataset(data_loader, len_limit)
        dist.destroy_process_group()



import numpy as np
from aperturedb import CSVParser
import logging
import os

logger = logging.getLogger(__name__)

HEADER_PATH = "filename"
HEADER_INDEX = "index"
HEADER_SET = "set"
HEADER_LABEL = "label"
PROPERTIES = "properties"
CONSTRAINTS = "constraints"


class DescriptorDataCSV(CSVParser.CSVParser):
    """
    **ApertureDB Descriptor Data.**

    This class loads the Descriptor Data which is present in a CSV file,
    and converts it into a series of ApertureDB queries.

    :::note Is backed by a CSV file with the following columns, and a NumPy array file "npz" for the descriptors:
    ``filename``, ``index``, ``set``, ``label``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP_NAME_N``
    :::

    **filename**: Path to a npz file which comprises of np arrays.

    **index**: The 0 based index of a np array in the npz file.

    **set**: The search space to restrict the knn search queries to.

    **label**: Arbitrary name given to the label associated with this descriptor.

    **PROP_NAME_1 .. PROP_NAME_N**: Arbitrarily assigned properties to this descriptor.

    **constraint_PROP_NAME_1**: A constraint to ensure uniqueness when inserting this descriptor.

    Example CSV file::

        filename,index,set,label,isTable,UUID,constraint_UUID
        /mnt/data/embeddings/kitchen.npz,0,kitchen,kitchen_table,True,AID-0X3E,AID-0X3E
        /mnt/data/embeddings/kitchen.npz,1,kitchen,kitchen_table,True,BXY-AB1Z,BXY-AB1Z
        /mnt/data/embeddings/dining_chairs.npz,1,dining_chairs,special_chair,False,COO-SE1R,COO-SE1R
        ...

    Example usage:

    ``` python

        data = DescriptorDataCSV("/path/to/DescriptorData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    ```



    :::info
    In the above example, the index uniquely identifies the actual np array from the many arrays in the npz file
    which is same for line 1 and line 2. The UUID and constraint_UUID ensure that a Descriptor is inserted only once in the DB.

    Association of an entity to a Descriptor can be specified by first ingesting other Objects, then Descriptors and finally by
    using [ConnectionDataCSV](/python_sdk/data_loaders/csv_wrappers/ConnectionDataCSV)

    In the above example, the constraint_UUID ensures that a connection with the specified
    UUID would be only inserted if it does not already exist in the database.
    :::

    """

    def __init__(self, filename: str, **kwargs):

        super().__init__(filename, **kwargs)
        self.npy_arrays = {}
        self.has_label = False

        self.props_keys = [x for x in self.header[3:]
                           if not x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[3:]
                                 if x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
        self.command = "AddDescriptor"

    def get_indices(self):
        return {
            "entity": {
                "_Descriptor": self.get_indexed_properties()
            }
        }

    def getitem(self, idx):
        idx = self.df.index.start + idx
        filename = os.path.join(self.relative_path_prefix,
                                self.df.loc[idx, HEADER_PATH])
        index = self.df.loc[idx, HEADER_INDEX]
        desc_set = self.df.loc[idx, HEADER_SET]

        descriptor, desc_ok = self.load_descriptor(filename, index)

        if not desc_ok:
            logger.error("Error loading descriptor: " + filename + ":" + index)
            raise Exception("Error loading descriptor: " +
                            filename + ":" + index)

        q = []
        blobs = []
        # for data in image_data:
        custom_fields = {"set": desc_set}
        if self.has_label:
            custom_fields["label"] = self.df.loc[idx, HEADER_LABEL]

        ad = self._basic_command(idx, custom_fields)

        blobs.append(descriptor)
        q.append(ad)

        return q, blobs

    # This function can be re-defined by user when they have
    # app-specific structure on the npz file.
    def retrieve_by_index(self, filename, index):

        try:
            desc = self.npy_arrays[filename][index]
        except:
            err_msg = "Cannot retrieve descriptor {} from {}".format(
                str(index), filename)
            raise Exception(err_msg)

        return desc

    def load_descriptor(self, filename, index):

        if filename not in self.npy_arrays:
            self.npy_arrays[filename] = np.load(filename)

        # Can be defined by the user.
        descriptor = self.retrieve_by_index(filename, index)

        if len(descriptor) < 0:
            return [], False

        descriptor = descriptor.astype('float32').tobytes()

        return descriptor, True

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[3] == HEADER_LABEL:
            self.has_label = True

        if self.header[0] != HEADER_PATH:
            raise Exception(
                "Error with CSV file field: filename. Must be first field")
        if self.header[1] != HEADER_INDEX:
            raise Exception(
                "Error with CSV file field: index. Must be second field")
        if self.header[2] != HEADER_SET:
            raise Exception(
                "Error with CSV file field: set. Must be third field")



import logging
from aperturedb.CSVParser import CSVParser, CONSTRAINTS_PREFIX
from aperturedb.Query import QueryBuilder, ObjectType

logger = logging.getLogger(__name__)

CONNECTION_CLASS = "ConnectionClass"
PROPERTIES = "properties"
CONSTRAINTS = "constraints"


class ConnectionDataCSV(CSVParser):
    """**ApertureDB Connection Data.**

    This class loads the Connection Data which is present in a CSV file,
    and converts it into a series of aperturedb queries.

    :::note Is backed by a CSV file with the following columns:
    ``ConnectionClass``, ``Class1@PROP_NAME``, ``Class2@PROP_NAME`` ... ``PROP_NAME_N``, ``constraint_PROP1``
    :::

    Example CSV file::

        ConnectionClass,_Image@id,_Descriptor@UUID,confidence,id,constraint_id
        has_image,321423532,AID-0X3E,0.4354,5432543254,5432543254
        has_image,42342522,BXY-AB1Z,0.6432,98476542,98476542
        ...

    **ConnectionClass**: Arbitrary class name for the entity this would be saved as.

    **``ClassName``@``PropertyName``**: This is a special combination of Class Name and Property Name that can uniquely identify an entity. ‘@’ is a delimiter, so should not be used in a property name.

    **PROP_NAME_1 .. PROP_NAME_N**: Arbitrary property names.

    **constraint_PROP_NAME_1**: A equality check against a unique property to ensure duplicates are not inserted.

    Example usage:

    ``` python

        data = ConnectionDataCSV("/path/to/ConnectionData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    ```


    :::info
    This example CSV's first row creates connection between an Image(id=321423532) and a descriptor(UUID=AID-0X3E)
    It also connects the images and descriptors in the subsequent rows.

    In the above example, the constraint_id ensures that a connection with the specified
    id would be only inserted if it does not already exist in the database.
    :::
    """

    def __init__(self, filename: str, **kwargs):
        super().__init__(filename, **kwargs)

        self.props_keys = [x for x in self.header[3:]
                           if not x.startswith(CONSTRAINTS_PREFIX)]

        self.constraints_keys = [x for x in self.header[3:]
                                 if x.startswith(CONSTRAINTS_PREFIX)]

        self.src_class = self.header[1].split("@")[0]
        self.src_key = self.header[1].split("@")[1]
        self.dst_class = self.header[2].split("@")[0]
        # Pandas appends a .n to the column name if there is a duplicate
        self.dst_key = self.header[2].split("@")[1].split(".")[0]
        self.command = "AddConnection"

    def get_indices(self):
        return {
            "connection": {
                cls: self.get_indexed_properties() for cls in self.df[CONNECTION_CLASS].unique()
            }
        }

    def getitem(self, idx):
        src_value = self.df.loc[idx, self.header[1]]
        dst_value = self.df.loc[idx, self.header[2]]
        connection_class = self.df.loc[idx, CONNECTION_CLASS]
        q = []
        members = ["_Image", "_Blob", "_Video", "_Descriptor"]

        try:
            ref_src = 1
            cmd_params = {
                "_ref": ref_src,
                "unique": True,
                "constraints": {
                    self.src_key: ["==", src_value]
                }
            }
            # Special case for objects with blobs
            if self.src_class in members:
                cmd_params["blobs"] = False
            q.append(QueryBuilder.find_command(self.src_class, cmd_params))

            ref_dst = 2
            cmd_params = {
                "_ref": ref_dst,
                "unique": True,
                "constraints": {
                    self.dst_key: ["==", dst_value]
                }
            }
            # Special case for objects with blobs
            if self.dst_class in members:
                cmd_params["blobs"] = False
            q.append(QueryBuilder.find_command(self.dst_class, cmd_params))

            ae = self._basic_command(idx,
                                     custom_fields={
                                         "class": connection_class,
                                         "src": ref_src,
                                         "dst": ref_dst,
                                     })

            q.append(ae)
        except KeyError as ex:
            error_str = " expected key='" + ex.args[0] + "' in '" + self.df.loc[idx] + \
                "'. Ignored."
            logger.error(error_str)

        return q, []

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] != CONNECTION_CLASS:
            raise Exception(
                "Error with CSV file field: 0 - " + CONNECTION_CLASS)
        if "@" not in self.header[1]:
            raise Exception("Error with CSV file field: 1")
        if "@" not in self.header[2]:
            raise Exception("Error with CSV file field: 2")



class TestUtils():

    def test_remove_all_objects(self, utils):
        assert utils.remove_all_objects() == True, \
            "Failed to remove all objects"

    def test_remove_all_indexes(self, utils):
        assert utils.remove_all_indexes() == True, \
            "Failed to remove all indexes"

    def test_get_descriptorset_list(self, utils):
        assert utils.get_descriptorset_list() == []



import cv2
import logging
import os
from aperturedb import CSVParser
from aperturedb.Sources import Sources
import boto3


logger = logging.getLogger(__name__)

HEADER_PATH = "filename"
PROPERTIES = "properties"
CONSTRAINTS = "constraints"
HEADER_URL = "url"
HEADER_S3_URL = "s3_url"
HEADER_GS_URL = "gs_url"


class VideoDataCSV(CSVParser.CSVParser):
    """
    **ApertureDB Video Data.**

    This class loads the Video Data which is present in a CSV file,
    and converts it into a series of ApertureDB queries.


    :::note Is backed by a CSV file with the following columns:

    ``filename``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

    OR

    ``url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

    OR

    ``s3_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``

    OR

    ``gs_url``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``, ``format``
    :::

    Example CSV file::

        filename,id,label,constraint_id
        /home/user/file1.mp4,321423532,dog,321423532
        /home/user/file2.mp4,42342522,cat,4234252
        ...

    Example usage:

    ``` python

        data = VideoDataCSV("/path/to/VideoData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    ```

    or

    ``` bash

        adb ingest from-csv /path/to/VideoData.csv --ingest-type VIDEO
    ```


    :::info
    In the above example, the constraint_id ensures that an Video with the specified
    id would be only inserted if it does not already exist in the database.
    :::
    """

    def __init__(self, filename: str, check_video: bool = True, **kwargs):
        self.source_types = [HEADER_PATH,
                             HEADER_URL, HEADER_S3_URL, HEADER_GS_URL]
        super().__init__(filename, **kwargs)
        self.check_video = check_video

        self.props_keys = [x for x in self.header[1:]
                           if not x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
        self.command = "AddVideo"
        self.loaders = {
            HEADER_PATH: self.load_video,
            HEADER_URL: self.load_url,
            HEADER_S3_URL: self.load_s3_url,
            HEADER_GS_URL: self.load_gs_url
        }
        self.source_type = self.header[0]
        if self.use_dask == False and self.source_type == HEADER_S3_URL:
            s3_client = boto3.client('s3')
            self.sources = Sources(3, s3_client=s3_client)
        else:
            self.sources = Sources(3)

    def get_indices(self):
        return {
            "entity": {
                "_Video": self.get_indexed_properties()
            }
        }

    def check_video_buf(self, video) -> bool:
        # check if video is valid
        print(len(video))
        return True

    def getitem(self, idx):
        self.relative_path_prefix = os.path.dirname(self.filename) \
            if self.source_type == HEADER_PATH and self.blobs_relative_to_csv else ""
        uri = os.path.join(self.relative_path_prefix,
                           self.df.loc[idx, self.source_type])
        video_ok, video = self.loaders[self.source_type](uri)

        if not video_ok:
            logger.error("Error loading video: " + uri)
            raise Exception("Error loading video: " + uri)

        q = []
        blobs = []
        av = self._basic_command(idx)
        blobs.append(video)
        q.append(av)

        return q, blobs

    def load_video(self, filename):

        if self.check_video:
            try:
                a = cv2.VideoCapture(filename)
                if a.isOpened() == False:
                    logger.error(f"Video reading Error: {filename}")
            except Exception as e:
                logger.error(f"Video Error: {filename}")
                logger.exception(e)

        try:
            fd = open(filename, "rb")
            buff = fd.read()
            fd.close()
            return True, buff
        except Exception as e:
            logger.error(f"Video Error: {filename}")
            logger.exception(e)

        return False, None

    def load_url(self, url):
        return self.sources.load_from_http_url(url, self.check_video_buf)

    def load_s3_url(self, s3_url):
        return self.sources.load_from_s3_url(s3_url, self.check_video_buf)

    def load_gs_url(self, gs_url):
        return self.sources.load_from_gs_url(gs_url, self.check_video_buf)

    def validate(self):

        self.header = list(self.df.columns.values)

        if self.header[0] not in self.source_types:
            field = self.header[0]
            allowed = ", ".join(self.source_types)
            raise Exception(
                f"Error with CSV file field: {field}. Allowed values: {allowed}")



from aperturedb.Subscriptable import Subscriptable
from aperturedb.transformers.transformer import Transformer
from PIL import Image
import io
import time
from .facenet import generate_embedding
from aperturedb.Utils import create_connector, Utils


class FacenetPyTorchEmbeddings(Transformer):
    """
    Generates the embeddings for the images using the Facenet Pytorch model.
    """

    def __init__(self, data: Subscriptable, **kwargs) -> None:
        """
        Args:
            data: Subscriptable object
            search_set_name: Name of the [descriptorset](/query_language/Reference/descriptor_commands/desc_commands/AddDescriptor) to use for the search.
        """
        super().__init__(data)
        self.search_set_name = kwargs.get(
            "search_set_name", "facenet_pytorch_embeddings")

        # Let's sample some data to figure out the descriptorset we need.
        if len(self._add_image_index) > 0:
            sample = self._get_embedding_from_blob(self.data[0][1][0])
            utils = Utils(create_connector())
            utils.add_descriptorset(self.search_set_name, dim=len(sample) // 4)

    def _get_embedding_from_blob(self, image_blob: bytes):
        pil_image = Image.open(io.BytesIO(image_blob))
        embedding = generate_embedding(pil_image)
        serialized = embedding.cpu().detach().numpy().tobytes()
        return serialized

    def getitem(self, subscript):
        start = time.time()
        self.ncalls += 1
        x = self.data[subscript]

        for ic in self._add_image_index:
            serialized = self._get_embedding_from_blob(
                x[1][self._add_image_index.index(ic)])
            x[1].append(serialized)
            x[0].append(
                {
                    "AddDescriptor": {
                        "set": self.search_set_name,
                        "connect": {
                            "ref": x[0][ic]["AddImage"]["_ref"]
                        }
                    }
                })
        self.cumulative_time += time.time() - start
        return x



import json

from aperturedb import CSVParser

HEADER_POLYGONS = "polygons"
IMG_KEY_PROP = "img_key_prop"
IMG_KEY_VAL = "img_key_value"
POLYGON_FIELDS = {
    "_label": "label",
}


class PolygonDataCSV(CSVParser.CSVParser):
    """
    **ApertureDB Polygon Data.**

    This class loads the Polygon Data which is present in a CSV file,
    and converts it into a series of ApertureDB queries.

    :::note Is backed by a CSV file with the following columns:
    ``IMG_KEY``, [``POLYGON_PROPERTY_1``, ... ``POLYGON_PROPERTY_N``,] [``constraint_POLYGON_PROPERTY_1``, ... ``constraint_POLYGON_PROPERTY_N``,] [``_label``,] ``polygons``
    :::

    **IMG_KEY**: identifies the name of the image property that will identify the
    image with which to associate each polygon object. This property should reliably
    identify at most a single image, like a unique id. The value in each row will be
    used to look up the image to which the polygon will attach.

    **POLYGON_PROPERTY_I**: declares the name of a property that will be assigned to all polygon objects. Any number of properties can be declared in this way.

    **constraint_POLYGON_PROPERTY_I**: declares that POLYGON_PROPERTY_I should be unique, and that a new polygon will not be added if there already exists one with the same value for this property. For each row, the value in this column should match the value in column POLYGON_PROPERTY_I.

    **_label**: optionally applies a label to the polygon objects.

    **polygons**: a JSON array of polygon regions. Each polygon region is itself an array of [x,y] vertices that describe the boundary of a single contiguous polygon. See also [Polygon API parameter](/query_language/Reference/shared_command_parameters/polygons).

    Example CSV file::

        image_id,polygon_id,constraint_polygon_id,category_id,_label,polygons
        397133,82445,82445,44,bottle,"[[[224.24, 297.18], [228.29, 297.18], ...]]"
        397133,119568,119568,67,dining table,"[[[292.37, 425.1], [340.6, 373.86], ...]]"
        ...

    Example usage:

    ``` python

        data = PolygonDataCSV("/path/to/PolygonData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    ```

    """

    def __init__(self, filename: str, **kwargs):

        super().__init__(filename, kwargs=kwargs)

        self.props_keys = []
        self.constraints_keys = []
        self.polygon_keys = []
        for key in self.header[1:-1]:
            if key in POLYGON_FIELDS.keys():
                self.polygon_keys.append(key)
            elif key.startswith(CSVParser.CONSTRAINTS_PREFIX):
                self.constraints_keys.append(key)
            else:
                self.props_keys.append(key)

        self.img_key = self.header[0]
        self.command = "AddPolygon"

    def get_indices(self):
        return {
            "entity": {
                "_Polygon": self.get_indexed_properties()
            }
        }

    def getitem(self, idx):
        idx = self.df.index.start + idx

        q = []

        img_id = self.df.loc[idx, self.img_key]

        fi = {
            "FindImage": {
                "_ref": 1,
                "constraints": {
                    self.img_key: ["==", img_id],
                },
                "blobs": False,
            },
        }
        q.append(fi)

        polygon_fields = {
            "image_ref": 1,
            "polygons": json.loads(self.df.loc[idx, HEADER_POLYGONS])
        }
        for key in self.polygon_keys:
            polygon_fields[POLYGON_FIELDS[key]] = self.df.loc[idx, key]

        ap = self._basic_command(idx, polygon_fields)
        q.append(ap)

        return q, []

    def validate(self):

        self.header = list(self.df.columns.values)

        if len(self.header) < 2:
            raise Exception(
                "Error with CSV file: must have at least two columns")
        if self.header[-1] != HEADER_POLYGONS:
            raise Exception("Error with CSV file field: " + HEADER_POLYGONS)




from aperturedb import ParallelQuery
from aperturedb.Connector import Connector

import numpy as np
import logging
logger = logging.getLogger(__name__)


class ParallelLoader(ParallelQuery.ParallelQuery):
    """
    **Parallel and Batch Loader for ApertureDB**

    This takes a dataset (which is a collection of homogeneous objects)
    or a derived class, and optimally inserts them into database by splitting them
    into batches, and passing the batches to multiple workers.
    """

    def __init__(self, db: Connector, dry_run: bool = False):
        super().__init__(db, dry_run=dry_run)
        self.type = "element"

    def get_existing_indices(self):
        schema_result, _ = self.db.query([{"GetSchema": {}}])
        schema = schema_result[0]["GetSchema"]
        existing_indices = {}
        if schema:
            for index_type in (("entity", "entities"), ("connection", "connections")):
                foo = schema.get(index_type[1]) or {}
                bar = foo.get("classes") or {}
                for cls_name, cls_schema in bar.items():
                    props = cls_schema.get("properties") or {}
                    for prop_name, prop_schema in props.items():
                        if prop_schema[1]:  # indicates property has an index
                            existing_indices.setdefault(index_type[0], {}).setdefault(
                                cls_name, set()).add(prop_name)
        return existing_indices

    def create_indices(self, indices) -> None:
        if len(indices) == 0:
            return

        existing_indices = self.get_existing_indices()
        new_indices = []
        for tp, classes in indices.items():
            ex_classes = existing_indices.get(tp, {})
            for cls, props in classes.items():
                ex_props = ex_classes.get(cls, set())
                for prop in props - ex_props:
                    new_indices.append({
                        "index_type": tp,
                        "class": cls,
                        "property_key": prop
                    })

        if len(new_indices) == 0:
            return

        logger.info(
            f"Creating {len(new_indices)} indices: {new_indices}.")

        create_indices = [{"CreateIndex": idx} for idx in new_indices]

        res, _ = self.db.query(
            create_indices)

        if self.db.check_status(res) != 0:
            logger.warning(
                "Failed to create indices; ingestion will be slow.")
            logger.warning(res)

    def query_setup(self, generator) -> None:
        if hasattr(generator, "get_indices"):
            self.create_indices(generator.get_indices())

    def ingest(self, generator, batchsize: int = 1, numthreads: int = 4, stats: bool = False) -> None:
        """
        **Method to ingest data into the database**

        Args:
            generator (_type_): The list of data, or a class derived from [Subscriptable](/python_sdk/helpers/Subscriptable) to be ingested.
            batchsize (int, optional): The size of batch to be ussed. Defaults to 1.
            numthreads (int, optional): Number of workers to create. Defaults to 4.
            stats (bool, optional): If stats need to be presented, realtime. Defaults to False.
        """
        logger.info(
            f"Starting ingestion with batchsize={batchsize}, numthreads={numthreads}")
        self.query(generator, batchsize, numthreads, stats)

    def print_stats(self) -> None:

        times = np.array(self.get_times())
        total_queries_exec = len(times)

        print("============ ApertureDB Loader Stats ============")
        print(f"Total time (s): {self.total_actions_time}")
        print(f"Total queries executed: {total_queries_exec}")

        succeeded_queries = sum([stat["succeeded_queries"]
                                for stat in self.actual_stats])
        succeeded_commands = sum([stat["succeeded_commands"]
                                  for stat in self.actual_stats])

        if succeeded_queries == 0:
            print("All queries failed!")

        else:
            mean = np.mean(times)
            std = np.std(times)
            tp = 0 if mean == 0 else 1 / mean * self.numthreads

            print(f"Avg Query time (s): {mean}")
            print(f"Query time std: {std}")
            print(f"Avg Query Throughput (q/s): {tp}")

            i_tp = self.total_actions / self.total_actions_time
            print(
                f"Overall insertion throughput ({self.type}/s): {i_tp if self.error_counter == 0 else 'NaN'}")

            if self.error_counter > 0:
                err_perc = 100 * self.error_counter / total_queries_exec
                print(f"Total errors encountered: {self.error_counter}")
                print(f"Errors (%): {err_perc}")

            # TODO this does not take into account that the last
            # batch may be smaller than batchsize
            print(f"Total inserted elements: {succeeded_queries}")
            print(f"Total successful commands: {succeeded_commands}")
        print("=================================================")



from aperturedb.transformers.transformer import Transformer
from aperturedb.Subscriptable import Subscriptable
from aperturedb.Utils import Utils, create_connector

from PIL import Image
import io
import logging
import uuid
import hashlib

logger = logging.getLogger(__name__)


class ImageProperties(Transformer):
    """
    This computes some image properties and adds them to the metadata.
    """

    def __init__(self, data: Subscriptable, **kwargs) -> None:
        super().__init__(data, **kwargs)
        utils = Utils(create_connector())

        if "adb_data_source" not in utils.get_indexed_props("_Image"):
            utils.create_entity_index("_Image", "adb_data_source")

    def getitem(self, subscript):
        x = self.data[subscript]
        try:
            # x is a transaction that has an add_image command and a blob
            for ic in self._add_image_index:
                blob_index = self._add_image_index.index(ic)
                src_properties = x[0][ic]["AddImage"]["properties"]
                # Compute the dynamic properties and apply them to metadata
                src_properties["adb_image_size"] = len(x[1][blob_index])
                src_properties["adb_image_sha256"] = hashlib.sha256(
                    x[1][blob_index]).hexdigest()

                # Compute the image dimensions.
                pil_image = Image.open(io.BytesIO(x[1][blob_index]))
                src_properties["adb_image_width"] = pil_image.width
                src_properties["adb_image_height"] = pil_image.height
                src_properties["adb_image_id"] = str(
                    src_properties["id"] if "id" in src_properties else uuid.uuid4().hex)

        except Exception as e:
            # Importantly, do not raise an exception here, since it will kill ingestion.
            # Create a log message instead, for post-mortem analysis.
            logger.exception(e.with_traceback(None), stack_info=True)

        return x



from collections import namedtuple
import os
from typing import List, Tuple
import zipfile
import pandas as pd
from unittest.mock import MagicMock, patch
from kaggle.api.kaggle_api_extended import KaggleApi
from aperturedb.KaggleData import KaggleData


kf = namedtuple("files", ["files"])
file = namedtuple("file", ["fileType", "totalBytes", "name"])


def custom_download(workdir, files):
    def dload_zip(dataset, path, quiet, unzip):
        assert(dataset in workdir)
        assert(path == workdir)
        assert(not unzip)
        assert(not quiet)
        # Simulate a download from kaggle.
        # the kaggle API leads to a zipped dataset downloaded into the dataset_ref path.
        os.makedirs(workdir, exist_ok=True)
        with zipfile.ZipFile(os.path.join(workdir, "file.zip"), "w") as zip:
            for file in files:
                zip.writestr(file["name"], file["contents"])
    return dload_zip


class GoodGuysBadGuysImageDataKaggle(KaggleData):
    def __init__(self):
        super().__init__(dataset_ref="gpiosenka/good-guysbad-guys-image-data-set")

    def generate_index(self, root: str, records_count: int) -> pd.DataFrame:
        return pd.read_csv(os.path.join(root, "index.csv"))

    def generate_query(self, idx: int) -> Tuple[List[dict], List[bytes]]:
        return super().generate_query(idx)


class CatDataKaggle(KaggleData):
    def __init__(self):
        super().__init__(dataset_ref="crawford/cat-dataset")

    def generate_index(self, root: str, records_count: int) -> pd.DataFrame:
        related_files = []
        for path, b, c in os.walk(root):
            for f in c:
                if f.endswith("jpg"):
                    related_files.append({
                        "image": os.path.join(path, f),
                        "annotation": os.path.join(path, f"{f}.cat")
                    })
        return pd.json_normalize(related_files)


class TestKaggleIngest():
    def common_setup(self) -> None:
        self.root = "kaggleds"
        self.authenticates = False

    def does_auth(self):
        # When kaggle.authenticate would be invoked.
        self.authenticates = True

    def test_builtin_indexer_csv(self):
        """
        This type of dataset comes wiht 1 CSV file, and Kaggle Dataset is able to
        process it without the need for a custom indexer.
        """
        self.common_setup()
        dataset_ref = "gpiosenka/good-guysbad-guys-image-data-set"
        f = file(".csv", 1024, "index.csv")
        k = kf(files = [f])
        archive_files = [{"name": "index.csv",
                          "contents": "filename,class\r\n001.jpg,1"}]
        dz = custom_download(f"kaggleds/{dataset_ref}", files=archive_files)

        # Mock the kaggle methods that are expected to be called from Kaggle Dataset.
        with patch.multiple(KaggleApi,
                            dataset_list_files = MagicMock(return_value=k),
                            dataset_download_files = MagicMock(side_effect=dz),
                            authenticate = MagicMock(
                                side_effect = self.does_auth)
                            ) as mocks:
            dataset = GoodGuysBadGuysImageDataKaggle()
        assert len(dataset) == 1
        assert os.path.exists(
            os.path.join(self.root, dataset_ref, "file.zip")) == True
        assert self.authenticates == True

    def test_ingest_annotations_images_without_index(self):
        self.common_setup()
        dataset_ref = "crawford/cat-dataset"
        f = file(".csv", 1024, "index.csv")
        k = kf(files = [f])
        archive_files = [
            {"name": "cat1.jpg", "contents": "filename,class\r\n001.jpg,1"},
            {"name": "cat1.jpg.cat", "contents": "1 2 3 4"},
            {"name": "cat2.jpg", "contents": "filename,class\r\n001.jpg,1"},
            {"name": "cat2.jpg.cat", "contents": "5 6 7 8"}
        ]
        dz = custom_download(f"kaggleds/{dataset_ref}", files=archive_files)

        # Mock the kaggle methods that are expected to be called from Kaggle Dataset.
        with patch.multiple(KaggleApi,
                            dataset_list_files = MagicMock(return_value=k),
                            dataset_download_files = MagicMock(side_effect=dz),
                            authenticate = MagicMock(
                                side_effect = self.does_auth)
                            ) as mocks:

            dataset = CatDataKaggle()
        assert len(dataset) == 2
        assert os.path.exists(
            os.path.join(self.root, dataset_ref, "file.zip")) == True
        assert self.authenticates == True



import json
import logging
import os
import shutil
import tempfile
from unittest.mock import MagicMock, patch

import typer
from typer.testing import CliRunner

from aperturedb.cli.configure import app

logger = logging.getLogger(__name__)


class TestConfigure():
    def check_contents(self, filename: str, section: str, config_name: str):
        with open(filename) as instream:
            obj = json.loads(instream.read())
            if section is not None:
                assert section in obj
                assert config_name in obj[section]
            else:
                assert config_name in obj

    def test_create_no_folder_exists_global(self):
        tmp = "/tmp/global"
        if os.path.exists(tmp):
            shutil.rmtree(tmp)
        fake_folder = tmp
        fake_file = os.path.join(tmp, "adb.json")
        with patch.multiple(typer,
                            get_app_dir = MagicMock(return_value = fake_folder)):
            runner = CliRunner()
            result = runner.invoke(app, ["create", "test"])
            assert result.exit_code == 0
            self.check_contents(fake_file, None, "test")

    def test_create_no_folder_exists_local(self):
        tmp = "/tmp/local"
        if os.path.exists(tmp):
            shutil.rmtree(tmp)
        fake_folder = tmp
        fake_file = os.path.join(tmp, ".aperturedb", "adb.json")
        with patch.multiple(os,
                            getcwd = MagicMock(return_value = fake_folder)):
            runner = CliRunner()
            result = runner.invoke(app, ["create", "test", "--no-as-global"])
            assert result.exit_code == 0
            self.check_contents(fake_file, None, "test")

    def test_create_folder_exists_global(self):
        with tempfile.TemporaryDirectory() as tmp:
            fake_folder = tmp
            fake_file = os.path.join(tmp, "adb.json")
            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = fake_folder)):
                runner = CliRunner()
                result = runner.invoke(app, ["create", "test"])
                assert result.exit_code == 0
                self.check_contents(fake_file, None, "test")

    def test_create_folder_exists_local(self):
        with tempfile.TemporaryDirectory() as tmp:
            fake_folder = tmp
            fake_file = os.path.join(tmp, ".aperturedb", "adb.json")
            with patch.multiple(os,
                                getcwd = MagicMock(return_value = fake_folder)):
                runner = CliRunner()
                result = runner.invoke(
                    app, ["create", "test", "--no-as-global"])
                assert result.exit_code == 0
                self.check_contents(fake_file, None, "test")

    def test_create_empty_file_exists_global(self):
        with tempfile.TemporaryDirectory() as tmp:
            fake_folder = tmp
            fake_file = os.path.join(tmp, "adb.json")
            with open(fake_file, "w") as outstream:
                pass
            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = fake_folder)):
                runner = CliRunner()
                result = runner.invoke(app, ["create", "test"])
                assert result.exit_code == 0
                self.check_contents(fake_file, None, "test")

    def test_create_empty_file_exists_local(self):
        with tempfile.TemporaryDirectory() as tmp:
            fake_folder = tmp
            fake_file = os.path.join(tmp, ".aperturedb", "adb.json")
            os.mkdir(os.path.join(fake_folder, ".aperturedb"))
            with open(fake_file, "w") as outstream:
                pass
            with patch.multiple(os,
                                getcwd = MagicMock(return_value = fake_folder)):
                runner = CliRunner()
                result = runner.invoke(
                    app, ["create", "test", "--no-as-global"])
                assert result.exit_code == 0
                self.check_contents(fake_file, None, "test")

    def test_ls_no_files(self):
        with tempfile.TemporaryDirectory() as tmp:
            fake_folder = tmp
            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = fake_folder)):
                with patch.multiple(os,
                                    getcwd = MagicMock(return_value = fake_folder)):
                    runner = CliRunner()
                    result = runner.invoke(app, ["ls"])
                    assert result.exit_code == 0
                    assert "adb is not configured yet." in result.stdout
                    assert "Please run adb config" in result.stdout

    def test_ls_empty_files(self):
        with tempfile.TemporaryDirectory() as tmp:
            fake_folder = tmp
            fake_file = os.path.join(tmp, "adb.json")
            with open(fake_file, "w") as outstream:
                pass
            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = fake_folder)):
                with patch.multiple(os,
                                    getcwd = MagicMock(return_value = fake_folder)):
                    runner = CliRunner()
                    result = runner.invoke(app, ["ls"])
                    assert "Failed to decode json" in result.stdout

    def test_ls_global_only(self, config):
        with tempfile.TemporaryDirectory() as tmp_global:
            fake_folder = tmp_global
            fake_file = os.path.join(tmp_global, "adb.json")
            with open(fake_file, "w") as outstream:
                outstream.write(json.dumps(config, indent=2))

            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = fake_folder)):

                with tempfile.TemporaryDirectory() as tmp_local:
                    with patch.multiple(os,
                                        getcwd = MagicMock(return_value = tmp_local)):

                        runner = CliRunner()
                        result = runner.invoke(app, ["ls"])
                        assert result.exit_code == 0

                        for key in config:
                            assert key in result.stdout

    def test_ls_local_only(self, config):
        with tempfile.TemporaryDirectory() as tmp_global:

            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = tmp_global)):

                with tempfile.TemporaryDirectory() as tmp_local:
                    fake_folder = tmp_local
                    fake_file = os.path.join(
                        tmp_local, ".aperturedb", "adb.json")
                    os.mkdir(os.path.join(fake_folder, ".aperturedb"))
                    with open(fake_file, "w") as outstream:
                        outstream.write(json.dumps(config, indent=2))
                    with patch.multiple(os,
                                        getcwd = MagicMock(return_value = tmp_local)):
                        runner = CliRunner()
                        result = runner.invoke(app, ["ls"])
                        assert result.exit_code == 0

                        for key in config:
                            assert key in result.stdout

    def test_ls_both(self, config):
        with tempfile.TemporaryDirectory() as tmp_global:
            fake_file = os.path.join(tmp_global, "adb.json")
            with open(fake_file, "w") as outstream:
                outstream.write(json.dumps(config, indent=2))

            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = tmp_global)):

                with tempfile.TemporaryDirectory() as tmp_local:
                    fake_folder = tmp_local
                    fake_file = os.path.join(
                        tmp_local, ".aperturedb", "adb.json")
                    os.mkdir(os.path.join(fake_folder, ".aperturedb"))
                    with open(fake_file, "w") as outstream:
                        outstream.write(json.dumps(config, indent=2))
                    with patch.multiple(os,
                                        getcwd = MagicMock(return_value = tmp_local)):
                        runner = CliRunner()
                        result = runner.invoke(app, ["ls"])
                        assert result.exit_code == 0

                        for key in config:
                            assert key in result.stdout

    def test_active_name_local_as_local(self, config):
        with tempfile.TemporaryDirectory() as tmp_global:
            fake_file = os.path.join(tmp_global, "adb.json")
            with open(fake_file, "w") as outstream:
                outstream.write(json.dumps(config, indent=2))

            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = tmp_global)):

                with tempfile.TemporaryDirectory() as tmp_local:
                    fake_folder = tmp_local
                    fake_file = os.path.join(
                        tmp_local, ".aperturedb", "adb.json")
                    os.mkdir(os.path.join(fake_folder, ".aperturedb"))
                    lc = {}
                    lc["second"] = config["first"]
                    lc["active"] = "second"
                    with open(fake_file, "w") as outstream:
                        outstream.write(json.dumps(lc, indent=2))
                    with patch.multiple(os,
                                        getcwd = MagicMock(return_value = tmp_local)):
                        runner = CliRunner()
                        result = runner.invoke(
                            app, ["activate", "second", "--no-as-global"])
                        assert result.exit_code == 0

    def test_active_name_global_as_global(self, config):
        with tempfile.TemporaryDirectory() as tmp_global:
            fake_file = os.path.join(tmp_global, "adb.json")
            with open(fake_file, "w") as outstream:
                outstream.write(json.dumps(config, indent=2))

            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = tmp_global)):

                with tempfile.TemporaryDirectory() as tmp_local:
                    fake_folder = tmp_local
                    fake_file = os.path.join(
                        tmp_local, ".aperturedb", "adb.json")
                    os.mkdir(os.path.join(fake_folder, ".aperturedb"))
                    with open(fake_file, "w") as outstream:
                        outstream.write(json.dumps(config, indent=2))
                    with patch.multiple(os,
                                        getcwd = MagicMock(return_value = tmp_local)):
                        runner = CliRunner()
                        result = runner.invoke(
                            app, ["activate", "first", "--as-global"])
                        assert result.exit_code == 0

    def test_active_name_local_as_global(self, config):
        with tempfile.TemporaryDirectory() as tmp_global:
            fake_file = os.path.join(tmp_global, "adb.json")
            with open(fake_file, "w") as outstream:
                outstream.write(json.dumps(config, indent=2))

            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = tmp_global)):

                with tempfile.TemporaryDirectory() as tmp_local:
                    fake_folder = tmp_local
                    fake_file = os.path.join(
                        tmp_local, ".aperturedb", "adb.json")
                    os.mkdir(os.path.join(fake_folder, ".aperturedb"))
                    lc = {}
                    lc["second"] = config["first"]
                    lc["active"] = "second"
                    with open(fake_file, "w") as outstream:
                        outstream.write(json.dumps(config, indent=2))
                    with patch.multiple(os,
                                        getcwd = MagicMock(return_value = tmp_local)):
                        runner = CliRunner()
                        result = runner.invoke(
                            app, ["activate", "second", "--as-global"])
                        assert result.exit_code == 2

    def test_active_name_global_as_local(self, config):
        with tempfile.TemporaryDirectory() as tmp_global:
            fake_file = os.path.join(tmp_global, "adb.json")
            with open(fake_file, "w") as outstream:
                outstream.write(json.dumps(config, indent=2))

            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = tmp_global)):

                with tempfile.TemporaryDirectory() as tmp_local:
                    fake_folder = tmp_local
                    fake_file = os.path.join(
                        tmp_local, ".aperturedb", "adb.json")
                    os.mkdir(os.path.join(fake_folder, ".aperturedb"))
                    lc = {}
                    lc["second"] = config["first"]
                    lc["active"] = "second"
                    with open(fake_file, "w") as outstream:
                        outstream.write(json.dumps(lc, indent=2))
                    with patch.multiple(os,
                                        getcwd = MagicMock(return_value = tmp_local)):
                        runner = CliRunner()
                        result = runner.invoke(app, ["activate", "first"])
                        assert result.exit_code == 0

    def test_active_name_not_present(self, config):
        with tempfile.TemporaryDirectory() as tmp_global:
            fake_file = os.path.join(tmp_global, "adb.json")
            with open(fake_file, "w") as outstream:
                outstream.write(json.dumps(config, indent=2))

            with patch.multiple(typer,
                                get_app_dir = MagicMock(return_value = tmp_global)):

                with tempfile.TemporaryDirectory() as tmp_local:
                    fake_folder = tmp_local
                    fake_file = os.path.join(
                        tmp_local, ".aperturedb", "adb.json")
                    os.mkdir(os.path.join(fake_folder, ".aperturedb"))
                    lc = {}
                    lc["second"] = config["first"]
                    with open(fake_file, "w") as outstream:
                        outstream.write(json.dumps(config, indent=2))
                    with patch.multiple(os,
                                        getcwd = MagicMock(return_value = tmp_local)):
                        runner = CliRunner()
                        result = runner.invoke(app, ["activate", "blah"])
                        assert result.exit_code == 2
                        assert "Configuration blah not found" in result.stdout



# This file only exists to support readable type hints

from typing import List, Dict, Any, Tuple

Command = Dict[str, Any]
Blob = bytes
Commands = List[Command]  # aka Query, but that's also a class name
Blobs = List[Blob]
CommandResponses = List[Dict]

Image = bytes
Video = bytes
Descriptor = bytes



from __future__ import annotations
import logging
from threading import Lock
import time
from types import SimpleNamespace
import dask
from dask.distributed import Client, LocalCluster, progress
from aperturedb.Connector import Connector

import multiprocessing as mp

from aperturedb.Stats import Stats

dask.config.set({"dataframe.convert-string": False})

logger = logging.getLogger(__name__)


class DaskManager:
    """
    **Class responsible for setting up a local cluster and assigning parts
    of data to each worker**
    """

    def __init__(self, num_workers: int = -1):
        self.__num_workers = num_workers
        # The -1 magic number is to use as many 90% of the cores (1 worker per core).
        # This can be overridden by the user.
        # Create a pool of workers.
        # TODO: see if the same pool can be reused for multiple tasks.
        workers = self.__num_workers if self.__num_workers != \
            -1 else int(0.9 * mp.cpu_count())

        self._cluster = LocalCluster(n_workers=workers)
        self._cluster.shutdown_on_close = False
        self._client = Client(self._cluster)
        dask.config.set(scheduler="distributed")

    def __del__(self):
        logger.info(".......Shutting cluster.........")
        self._client.close()
        self._cluster.close()

    def run(self, QueryClass: type[ParallelQuery], db: Connector, generator, batchsize, stats):
        def process(df, host, port, use_ssl, session, connnector_type):
            metrics = Stats()
            # Dask reads data in partitions, and the first partition is of 2 rows, with all
            # values as 'foo'. This is for sampling the column names and types. Should not process
            # those rows.
            if len(df) == 2:
                if (df.iloc[0, 0] == "a" and df.isna().iloc[1, 0]) or df.iloc[0, 0] == "foo":
                    return
            count = 0
            try:
                shared_data = SimpleNamespace()
                shared_data.session = session
                shared_data.lock = Lock()
                db = connnector_type(host=host, port=port,
                                     use_ssl=use_ssl, shared_data=shared_data)
            except Exception as e:
                logger.exception(e)
            #from aperturedb.ParallelLoader import ParallelLoader
            loader = QueryClass(db)
            for i in range(0, len(df), batchsize):
                end = min(i + batchsize, len(df))
                slice = df[i:end]
                data = generator.__class__(
                    filename=generator.filename,
                    df=slice,
                    blobs_relative_to_csv=generator.blobs_relative_to_csv)

                loader.query(generator=data, batchsize=len(
                    slice), numthreads=1, stats=False)
                count += 1
                metrics.times_arr.extend(loader.times_arr)
                metrics.error_counter += loader.error_counter
                metrics.succeeded_queries += loader.get_succeeded_queries()
                metrics.succeeded_commands += loader.get_succeeded_commands()

            return metrics

        start_time = time.time()
        # Passing DB as an argument to function is not supported by Dask,
        # so we pass session and host/port instead.
        computation = generator.df.map_partitions(
            process,
            db.host,
            db.port,
            db.use_ssl,
            db.shared_data.session,
            type(db))
        computation = computation.persist()
        if stats:
            progress(computation)
        results = computation.compute()

        return results, time.time() - start_time



#!/usr/bin/env python3
import logging
import datetime
import os
import json
import requests

logger = logging.getLogger(__name__)

__version__ = "0.4.24"

# set log level
logger.setLevel(logging.DEBUG)
formatter    = logging.Formatter(
    "%(asctime)s : %(levelname)s : %(name)s : %(thread)d : %(lineno)d : %(message)s")

log_file_level = logging.getLevelName(os.getenv("LOG_FILE_LEVEL", "WARN"))
log_console_level = logging.getLevelName(
    os.getenv("LOG_CONSOLE_LEVEL", "ERROR"))

# define file handler and set formatter
error_file_handler = logging.FileHandler(
    f"error.{datetime.datetime.now().isoformat()}.log", delay=True)
error_file_handler.setFormatter(formatter)
error_file_handler.setLevel(log_file_level)
logger.addHandler(error_file_handler)

error_console_handler = logging.StreamHandler()
error_console_handler.setLevel(log_console_level)
error_console_handler.setFormatter(formatter)
logger.addHandler(error_console_handler)

try:
    latest_version = json.loads(requests.get(
        "https://pypi.org/pypi/aperturedb/json").text)["info"]["version"]
except Exception as e:
    logger.warning(
        f"Failed to get latest version: {e}. You are using version {__version__}")
    latest_version = None
if __version__ != latest_version:
    logger.warning(
        f"The latest version of aperturedb is {latest_version}. You are using version {__version__}. It is recommended to upgrade.")



from aperturedb import CSVParser
from aperturedb.EntityUpdateDataCSV import SingleEntityUpdateDataCSV
import logging

logger = logging.getLogger(__name__)
ENTITY_CLASS = "EntityClass"
PROPERTIES = "properties"
CONSTRAINTS = "constraints"


class EntityDataCSV(CSVParser.CSVParser):
    """**ApertureDB Entity Data.**

    This class loads the Entity Data which is present in a CSV file,
    and converts it into a series of ApertureDB queries.

    :::note Is backed by a CSV file with the following columns:
    ``EntityClass``, ``PROP_NAME_1``, ... ``PROP_NAME_N``, ``constraint_PROP1``
    :::

    Example CSV file::

        EntityClass,name,lastname,age,id,constraint_id
        Person,John,Salchi,69,321423532,321423532
        Person,Johna,Salchi,63,42342522,42342522
        ...

    Example usage:

    ``` python

        data = EntityDataCSV("/path/to/EntityData.csv")
        loader = ParallelLoader(db)
        loader.ingest(data)
    ```


    :::info
    In the above example, the constraint_id ensures that a Entity with the specified
    id would be only inserted if it does not already exist in the database.
    :::

    """

    def __init__(self, filename: str, **kwargs):
        super().__init__(filename, **kwargs)

        self.props_keys = [x for x in self.header[1:]
                           if not x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
        self.constraints_keys = [x for x in self.header[1:]
                                 if x.startswith(CSVParser.CONSTRAINTS_PREFIX)]
        self.command = "AddEntity"

    def get_indices(self):
        return {
            "entity": {
                cls: self.get_indexed_properties() for cls in self.df[ENTITY_CLASS].unique()
            }
        }

    def getitem(self, idx):
        idx = self.df.index.start + idx
        eclass = self.df.loc[idx, ENTITY_CLASS]
        q = []
        ae = self._basic_command(idx,
                                 custom_fields={
                                     "class": eclass
                                 })

        q.append(ae)
        return q, []

    def validate(self):
        if self.header[0] != ENTITY_CLASS:
            raise Exception("Error with CSV file field: " + ENTITY_CLASS)

# Used when a csv has a single entity type that needs to be deleted


class EntityDeleteDataCSV(CSVParser.CSVParser):
    """**ApertureDB Entity Delete Data.**

    This class loads the Entity Data which is present in a CSV file,
    and converts it into a series of ApertureDB deletes.

    :::note::
    Expects a CSV file with the following columns:

        ``constraint_PROP1``
    :::

    Example CSV file::

        constraint_id
        321423532
        42342522
        ...

    Example usage:

   ```python

        data = ImageDeleteDataCSV("/path/to/UnusedImages.csv")
        loader = ParallelQuery(db)
        loader.query(data)
    ```


    :::info
    In the above example, the constraint_id ensures that a Entity with the specified
    id would be only deleted.

    Note that you can take a csv with normal prop data and this will ignore it, so you
    could use input to a loader to this.
    :::


    """

    def __init__(self, entity_class, filename, df=None, use_dask=False):
        super().__init__(filename, df=df, use_dask=use_dask)
        self.command = "Delete" + entity_class
        self.constraint_keyword = "constraints"
        if not use_dask:
            self.constraint_keys = [x for x in self.header[0:]]

    def getitem(self, idx):
        idx = self.df.index.start + idx
        q = []
        entity_delete = self._basic_command(idx)

        q.append(entity_delete)
        return q, []

    def validate(self):
        # all we require is a valid csv with 1 or more columns.
        return True


class ImageDeleteDataCSV(EntityDeleteDataCSV):
    """
    ***ApertureData CSV Loader class for deleting images***
    Usage details in EntityDeleteDataCSV
    """

    def __init__(self, filename, df=None, use_dask=False):
        super().__init__("Image", filename, df=df, use_dask=use_dask)



from itertools import combinations
import logging
import pytest
from aperturedb.EntityDataCSV import EntityDataCSV
from aperturedb.ParallelLoader import ParallelLoader
from aperturedb.ParallelQuery import ParallelQuery
from aperturedb.Subscriptable import Subscriptable
from aperturedb.Connector import Connector
import math

logger = logging.getLogger(__name__)


class EntityWithResponseDataCSV(EntityDataCSV):
    def __init__(self, filename, requests, responses):
        super().__init__(filename)
        self.requests = requests
        self.responses = responses

    def response_handler(self, request, input_blob, response, output_blob):
        self.requests.append(request)
        self.responses.append(response)


class GeneratorWithNCommands(Subscriptable):
    def __init__(self, commands_per_query=1, blobs_per_query=0, elements=100) -> None:
        super().__init__()
        self.commands_per_query = commands_per_query
        self.blobs_per_query = blobs_per_query
        self.elements = elements

    def __len__(self):
        return self.elements

    def getitem(self, subscript):
        query = []
        blobs = []
        for i in range(self.commands_per_query):
            query.append({
                "FindEntity": {

                }
            })
        buff = None
        with open(__file__, "rb") as ins:
            buff = ins.read()

        for i in range(self.blobs_per_query):
            blobs.append(buff)

        return query, blobs


class QGPersons(Subscriptable):
    def __init__(self, requests, responses, cpq, response_blobs) -> None:
        super().__init__()
        self.requests = requests
        self.responses = responses
        self.response_blobs = response_blobs
        self.cpq = cpq

    def __len__(self):
        return math.ceil(10 / self.cpq)

    def getitem(self, subscript):
        query = []
        for i in range(self.cpq):
            query.append(
                {
                    "FindEntity": {
                        "with_class": "Person",
                        "results": {
                            "all_properties": True
                        },
                        "constraints": {
                            "age": [">=", (subscript * self.cpq + i) * 10, "<", (subscript * self.cpq + 1 + i) * 10]
                        }
                    }
                }
            )
            if subscript * self.cpq + 1 + i == 10:
                break
        return query, []

    def response_handler(self, request, input_blob, response, output_blob):
        self.requests.append(request)
        self.responses.append(response)
        if output_blob is not None and len(output_blob) > 0:
            self.response_blobs.append(output_blob)


class QGImages(QGPersons):
    def getitem(self, subscript):
        query = []
        for i in range(self.cpq):
            q = {
                "FindImage": {

                }
            }
            query.append(q)
        return query, []


# Fake DB query.
def mock_query(self, request, blobs):
    self.response = 0
    return [], []

# Fake DB query. Depending in the transaction, return a response increasing number of images.
# Also add a item (int) to the blobs list for each image returned (to simulate the image data)
# The number of blobs returned is validated at the end.


def query_mocker_factory(response_blobs_count):
    def mock_query(self, request, blobs):
        self.response = 0
        mock_responses = [{
            "FindImage": {
                "returned": (i + 1),
                "status": 0
            }
        } for i, c in enumerate(request)]
        mock_blobs = [i for i in range(
            int(response_blobs_count * len(request)))]
        return mock_responses, mock_blobs
    return mock_query


class TestResponseHandler():
    def cleanDB(self, utils):
        logger.debug(f"Cleaning existing data")
        self.requests = []
        self.responses = []

    @pytest.mark.parametrize("records_count", [(1), (99), (100)])
    def test_response_handler(self, records_count, db, utils, monkeypatch):
        monkeypatch.setattr(Connector, "query", lambda s,
                            r, b: mock_query(s, r, b))
        self.cleanDB(utils=utils)
        logger.warning(f"{records_count}")
        persons = EntityWithResponseDataCSV(
            "./input/persons.adb.csv", self.requests, self.responses)
        persons = persons[:records_count]
        loader = ParallelLoader(db)
        loader.ingest(persons, batchsize=99,
                      numthreads=31,
                      stats=True)
        assert loader.error_counter == 0

        assert len(persons) == records_count
        logger.info(self.responses)
        assert len(self.responses) == records_count
        logger.info(self.requests)
        assert len(self.requests) == records_count

    @pytest.mark.parametrize("commands_blobs", combinations(range(1, 4), 2))
    def test_varying_commands_blobs(self, db, commands_blobs, monkeypatch):
        monkeypatch.setattr(Connector, "query", lambda s,
                            r, b: mock_query(s, r, b))
        i, j  = commands_blobs
        generator = GeneratorWithNCommands(i, j)
        querier = ParallelQuery(db, dry_run=True)
        querier.query(generator, batchsize=99,
                      numthreads=31,
                      stats=True)
        assert querier.commands_per_query == i
        assert querier.blobs_per_query == j

    @pytest.mark.parametrize("cpq", range(1, 11))
    def test_varying_response(self, db, utils, cpq):
        self.cleanDB(utils=utils)
        persons = EntityWithResponseDataCSV(
            "./input/persons.adb.csv", self.requests, self.responses)
        loader = ParallelLoader(db)
        loader.ingest(persons, batchsize=99,
                      numthreads=31,
                      stats=True)
        assert loader.error_counter == 0

        dist = persons.df.groupby(persons.df.age // 10 * 10).count()
        logger.debug(dist["age"])

        self.requests = []
        self.responses = []
        self.response_blobs = []
        generator = QGPersons(self.requests, self.responses,
                              cpq, self.response_blobs)
        querier = ParallelQuery(db)
        querier.query(generator, batchsize=100,
                      numthreads=1,
                      stats=True)
        dist_by_ages = {}
        for req, resp in zip(self.requests, self.responses):
            logger.debug(req)
            assert len(req) in [cpq, 10 % cpq]
            for rq, rr in zip(req, resp):
                for key in rq:
                    age_group = rq[key]["constraints"]["age"][1]
                    count = rr[key]["returned"]
                    dist_by_ages[age_group] = count
        logger.debug(dist_by_ages)
        for k in dist_by_ages:
            assert dist_by_ages[k] == dist["age"][k]

    def test_response_blobs(self, db, utils, monkeypatch):
        monkeypatch.setattr(Connector, "query", lambda s, r,
                            b: query_mocker_factory(5.5)(s, r, b))
        self.cleanDB(utils=utils)
        self.response_blobs = []
        generator = QGImages(self.requests, self.responses,
                             1, self.response_blobs)
        querier = ParallelQuery(db)
        querier.query(generator, batchsize=99,
                      numthreads=31,
                      stats=True)
        assert len(self.response_blobs) == len(self.responses)
        for i, resp in enumerate(self.responses):
            for key in resp[0]:
                assert resp[0][key]["returned"] == i + 1
                assert resp[0][key]["status"] == 0
                assert len(self.response_blobs[i]) == resp[0][key]["returned"]



from __future__ import annotations
from typing import Any, Dict, List, Union
from aperturedb.Query import Query, ObjectType

from aperturedb.Subscriptable import Subscriptable
from aperturedb.Constraints import Constraints
from aperturedb.Connector import Connector
from aperturedb.ParallelQuery import execute_batch
from aperturedb.Query import QueryBuilder
import pandas as pd


class Entities(Subscriptable):
    """
    This class is the common class to query any entity from ApertureDB.
    The specialized subclasses, which provide a more user friendly interface, are:
    * [Blobs](/python_sdk/object_wrappers/Blobs)
    * [Bounding Boxes](/python_sdk/object_wrappers/BoundingBoxes)
    * [Images](/python_sdk/object_wrappers/Images)
    * [Polygons](/python_sdk/object_wrappers/Polygons)
    * [Videos](/python_sdk/object_wrappers/Videos)
    """
    db_object = "Entity"
    find_command = f"Find{db_object}"
    update_command = f"Update{db_object}"

    @classmethod
    def retrieve_entities(cls,
                          db: Connector,
                          spec: Query,
                          with_adjacent: Dict[str, Query] = None
                          ) -> List[Entities]:
        """
        Using the Entities.retrieve method, is a simple layer, with typical native queries converted
        using :class:`~aperturedb.Query.Query`

        Args:
            db (Connector): _description_
            spec (Query): _description_

        Raises:
            e: _description_

        Returns:
            List[Entities]: _description_
        """

        # Since adjacent items are usually a way to filter the results,
        # the native query is constructed in the reverse order, with
        # first filtering out the relevant items based on adjacent items.
        fs = None
        count = 0
        if with_adjacent:
            for k, v in with_adjacent.items():
                if fs is None:
                    fs = v
                else:
                    fs = fs.connected_to(v)
                count += 1
            # Eventually, connect the specification of Images to the specification of the adjacent items.
            fs = fs.connected_to(spec)
        else:
            fs = spec

        spec = fs

        cls.known_entities = load_entities_registry(
            custom_entities=spec.command_properties(prop="with_class"))
        cls.db = db

        query = spec.query()
        print(f"query={query}")
        res, r, b = execute_batch(query, [], db)
        if res > 0:
            print(f"resp={r}")
        results = []
        for wc, req, blobs, resp in zip(
                spec.command_properties(prop="with_class"),
                spec.command_properties(prop="find_command"),
                spec.command_properties(prop="blobs"),
                r):
            subresponse = resp[req]['entities']
            if not isinstance(subresponse, list):
                flattened = []
                for previtem in results[-1]:
                    flattened.extend(subresponse[previtem["_uniqueid"]])
                subresponse = flattened
            try:
                entities = cls.known_entities[wc](
                    db=db, response=subresponse, type=wc, spec=spec)
                entities.blobs = blobs

                results.append(entities)
            except Exception as e:
                print(e)
                print(cls.known_entities)
                raise e

        cls.__postprocess__(entities=results[-1], with_adjacent=with_adjacent)
        return results

    @classmethod
    def retrieve(cls,
                 db: Connector,
                 spec: Query,
                 with_adjacent: Dict[str, Query] = None) -> Entities:
        spec.db_object = cls.db_object

        results = Entities.retrieve_entities(
            db=db, spec=spec, with_adjacent=with_adjacent)

        # This is a very naive assumption, we will stop querying once
        # the object of interest is the in the responses.
        objects = results[-1]

        return objects

    # This needs to be defined so that the application can access the adjacent items,
    # with every item of this iterable.

    @classmethod
    def __decorator(cls, index, adjacent):
        item = {}
        for k, v in adjacent.items():
            item[k] = v[index]
        return item

    @classmethod
    def __postprocess__(cls, entities: Entities, with_adjacent: object = None):
        adjacent = {}
        if with_adjacent:
            for k, v in with_adjacent.items():
                adjacent[k] = entities.get_connected_entities(
                    etype=v.with_class,
                    constraints=v.constraints)

            entities.decorator = cls.__decorator
            entities.adjacent = adjacent

    def __init__(self,
                 db: Connector = None,
                 response: dict = None,
                 type: str = None,
                 spec: Query = None) -> None:
        super().__init__()
        self.db = db
        self.response = response
        self.type = type
        self.decorator = None
        self.get_image = False
        self.spec = spec

    def getitem(self, idx):
        item = self.response[idx]
        if self.decorator is not None:
            for k, v in self.decorator(idx, self.adjacent).items():
                item[k] = v

        return item

    def __len__(self):
        return len(self.response)

    def filter(self, predicate):
        return self.known_entities[self.type](db=self.db, response=list(filter(predicate, self.response)), type=self.type)

    def __add__(self, other: Entities) -> Entities:
        return Entities(response = self.response + other.response, type=self.type)

    def __sub__(self, other: Entities) -> Entities:
        return Entities(response = [x for x in self.response if x not in other.response], type=self.type)

    def sort(self, key) -> Entities:
        return Entities(response = sorted(self.response, key=key), type=self.type)

    def inspect(self, **kwargs) -> pd.DataFrame:
        return pd.json_normalize([item for item in self])

    def update_properties(self, extra_properties: List[dict]) -> bool:
        for entity, properties in zip(self, extra_properties):
            query = [
                {
                    self.update_command: {
                        "constraints": {
                            "_uniqueid": ["==", entity["_uniqueid"]]
                        },
                        "properties": properties
                    }
                }
            ]
            res, r, b = execute_batch(query, [], self.db)

    def get_connected_entities(self,  etype: Union[ObjectType, str], constraints: Constraints = None) -> List[Entities]:
        """
        Gets all entities adjacent to and clustered around items of the collection

        Args:
            pk (str): _description_
            type (ObjectType): _description_
            constraints (Constraints, optional): _description_. Defaults to None.

        Returns:
            List[Entities]: _description_
        """
        result = []
        entity_class = etype.value if isinstance(etype, ObjectType) else etype
        for entity in self:
            params_src = {
                "_ref": 1,
                "unique": False,
                "constraints": {
                    "_uniqueid": ["==", entity["_uniqueid"]]
                }
            }
            params_dst = {
                "is_connected_to": {
                    "ref": 1
                },
                "with_class": entity_class,
                "constraints": constraints.constraints,
                "results": {
                    "all_properties": True
                }
            }

            query = [
                QueryBuilder.find_command(self.db_object, params=params_src),
                QueryBuilder.find_command(
                    oclass=entity_class, params=params_dst)
            ]
            res, r, b = execute_batch(query, [], self.db)
            cl = Entities
            if entity_class in self.known_entities:
                cl = self.known_entities[entity_class]
            result.append(cl(
                db=self.db, response=r[1]["FindEntity"]["entities"], type=entity_class))
        return result

    def get_blob(self, entity) -> Any:
        """
        Helper to get blobs for FindImage, FindVideo and FindBlob commands.
        """
        cmd_params = {
            "constraints": {
                "_uniqueid": ["==", entity["_uniqueid"]]
            },
            "blobs": True,
            "uniqueids": True,
            "results": {
                "count": True
            }
        }
        # An entities object is created with the response of the query
        # or a spec. Check if the spec is not None, and if it has operations.
        if self.spec and self.spec.operations:
            cmd_params["operations"] = self.spec.operations.operations_arr
        query = [
            QueryBuilder().find_command(self.db_object, params=cmd_params)
        ]
        res, r, b = execute_batch(query, [], self.db)
        return b[0]


def load_entities_registry(custom_entities: List[str] = None) -> dict:
    from aperturedb.Polygons import Polygons
    from aperturedb.Images import Images
    from aperturedb.Blobs import Blobs
    from aperturedb.BoundingBoxes import BoundingBoxes
    from aperturedb.Videos import Videos

    known_entities = {
        ObjectType.POLYGON.value: Polygons,
        ObjectType.IMAGE.value: Images,
        ObjectType.VIDEO.value: Videos,
        ObjectType.BOUNDING_BOX.value: BoundingBoxes,
        ObjectType.BLOB.value: Blobs,
    }
    for entity in set(custom_entities):
        if entity not in known_entities:
            known_entities[entity] = Entities
    return known_entities
